
from camagick.pipe.only import _FilterBase

class Processor(_FilterBase):
    '''
    Blacklist-based data blocker -- allows only named data to pass.

    Blocks data items with keys that match any
    of the regex expressions. The match is case-sensitive.
    '''
    def __init__(self, *items: list):
        super().__init__(False, *items)



from camagick.processor import (
    ProcessorBase, FlowBase, SourceBase, SinkBase,
    DuplicateDataTag
)



import asyncio, logging, pprint

logger = logging.getLogger(__name__)

class Processor(FlowBase):
    '''
    Parallel processing of data

    Pipes are run in asynchronously, in parallel. Data outputs of all the branches
    are joined together.

    CLI syntax is { ... } brackets (mind the spaces).
    '''
    
    def __init__(self, *pipes, **kw):
        super().__init__(*pipes)


    async def __call__(self, data=None, context=None):
        logger.debug(f'msg="Fanout" pipes={self.subpipes}')
        out_data = await asyncio.gather(*[p(data, context) for p in self.subpipes],
                                        return_exceptions=False)

        ret = {}
        for o,p in zip(out_data, self.subpipes):

            # Ignore empty results
            if o is None or len(o) == 0:
                continue

            # Don't merge results of sinks into the final data pack.
            # Push them into context instead (ACHTUNG, this might overwrite
            # context information).
            if isinstance(p, SinkBase):
                for k,v in o.items():
                    if k not in context:
                        context[k] = v
                continue

            # make sure we don't overwrite data
            for k in o:
                if k in out_data and out_data[k] is not None:
                    raise DuplicateDataTag(f'msg="Duplicate data tag" tag={k}')

            ret.update(o)

        return ret



from camagick.pipe.slice import Processor, _parse_slice

from numpy.random import rand as rand_array

async def test_slice():

    sobj = [
        [ slice(10, 15), slice(20,40,2), slice(0,50) ],
        [ None,          slice(20,40,2), slice(0,50) ],
        [ slice(10),     slice(20,40,2), slice(50) ],
        [ 10,            slice(20,40,2), slice(0,50) ],
        [ slice(10, 11), slice(20,40,2), slice(0,50) ],
    ]

    sstr = [
        '10:15,20:40:2,0:50',
        ',20:40:2,0:50',        
        ':10,20:40:2,:50',
        '10,20:40:2,0:50',        
        '10:11,20:40:2,0:50',
    ]

    inp = rand_array(17, 43, 59)

    for so,ss in zip(sobj,sstr):

        tmp = [
            _parse_slice(part) for part in ss.split(',')
        ]
        
        r1 = await Processor(*so)({'data': inp})
        r2 = await Processor(*tmp)({'data': inp})
        
        print(r1.keys(), r1['data'].shape, inp[*so].shape)
        print(ss)        
        print(so)
        print(tmp)

        for a,b in zip(r1.values(), r2.values()):
            assert ((a-b)**2).sum() == 0.0



from camagick.processor import ProcessorBase
import re

class Processor(ProcessorBase):    
    '''
    Renames data tags.
    '''

    def __init__(self, fmt: str, only: str = ''):
        '''
        Args:
            fmt: new tag format. Use `{}` or `{tag}` to fill in the
              old tag names, or any context key for the corresponding
              context data.

            only: if set to a non-empty string, only tags matching this
              regex pattern are renamed.
        '''

        self._format = fmt
        self._only = re.compile(only) if len(only)>0 else None

    
    async def __call__(self, data=None, context=None):

        if data is None:
            return None

        matcher = lambda x: \
            (self._only is None) or \
            (self._only.match(x[0]) is not None)

        output = {
            self._format.format(tag, tag=tag, **context):val \
            for tag, val in filter(matcher, data.items())
        }
        
        output.update({
            tag:val for tag, val in filter(lambda x: not matcher(x), data.items())
        })

        return output



from camagick.processor import (
    ProcessorBase, SourceBase, FlowBase, SinkBase,
    DuplicateDataTag
)

import asyncio, logging, pprint

logger = logging.getLogger(__name__)

class Processor(FlowBase):
    '''
    Sequential execution of processors

    For Pipe-type processours, output data from one processor is passed
    as input data to the next. The final data set is returned as a result
    of the chain.
    
    If a given processor is a Source-type, then its output data is rather
    merged into the existing data set instead of fully replacing it.
    '''

    async def __call__(self, data=None, context=None):
        inp = data
        for pipe in self.subpipes:
            outp = await pipe(inp, context)
            if outp is None:
                outp = {}
            logger.debug(f'msg="Chain element" pipe={pipe} '
                         f'in={inp.keys() if inp is not None else None} '
                         f'outp={outp.keys() if outp is not None else None}')

            if isinstance(pipe, SourceBase) and (inp is not None):
                for k,v in outp.items():
                    if k in inp:
                        raise DuplicateDataTag(f'msg="Duplicate tag" tag={k} pipe={pipe}')
                    inp[k] = v
            else:
                inp = outp

        return inp



import pytest, asyncio, time, traceback
import multiprocessing as mp

from camagick.source.scratch import Processor as GenSource
from camagick.source.ioc import Processor as IocSource
from camagick.source.epics import Processor as EpicsSource

from camagick.sink.ioc import Processor as IocSink
from camagick.sink.epics import Processor as EpicsSink
from camagick.sink.summary import Processor as SummarySink

from camagick.flow.chain import Processor as ChainFlow
from camagick.flow.fanout import Processor as FanoutFlow

from camagick.pipe.pace import Processor as Pace
from camagick.pipe.only import Processor as Only
from camagick.pipe.exclude import Processor as Exclude

from camagick.executor import PipeExecutor

from camagick.reader import AsyncReader

import logging

def ioc_main(chain):
    logger = logging.getLogger("camagick")
    logging.basicConfig()    
    logger.setLevel(logging.DEBUG)
    asyncio.run(PipeExecutor(chain).run())


@pytest.fixture(scope='module')
def ioc():

    prefix = 'tee:'
    
    chain = ChainFlow(
        IocSource(prefix=prefix, mark="mark"),

        IocSource(prefix=prefix, counter1="counter1(0)", hang=False),

        GenSource("random", iffy=1, jiffy=100),
        Only(".*iffy"),
        IocSink(prefix=prefix),

        GenSource("zeros", "idle"),
        EpicsSink(idle="tee:mark"),

        IocSource(prefix=prefix, counter2="counter2(0)", hang=False),

        Pace(Hz=1.0),
        GenSource("ones", "busy"),
        EpicsSink(busy="tee:mark")
    )
    
    p = mp.Process(target=ioc_main, args=[chain], kwargs={})
    p.daemon = True
    p.start()

    #time.sleep(1.0)

    yield {
        'prefix': prefix,
        'process': p
    }

    print('Terminating IOC')
    p.terminate()
             

async def test_reader_continuous(ioc):
    pref = ioc['prefix']
    t0 = time.time()
    data = {}

    readout_vars = (
        f'{pref}iffy',
        f'{pref}mark'
    )

    async with AsyncReader(*readout_vars, mode="continuous", dwell=0.0) as reader:
        while (time.time()-t0) < 5.0:
            d = await reader.wait_for_incoming()
            if len(d) == 0:
                await asyncio.sleep(0.1)
                continue

            print('received:', d)
            if 'tee:iffy' in d:
                data.setdefault(d['tee:iffy'], 0)
                data[d['tee:iffy']] += 1
            print(data, len(d))

            # We're not sure how many items we receive on the first run,
            # but come the 2nd run, there should be one full set.
            if time.time()-t0 > 1.0:
                assert len(d) == 2

            await asyncio.sleep(0.2)

    ## We should have received 5 distinct random numbers,
    ## and each of them several times over.
    assert len(data) == 5
    for k,v in data.items():
        assert v > 1



async def test_reader_hasty(ioc):
    pref = ioc['prefix']
    t0 = time.time()
    data = {}

    received_sum = 0
    loop_sum = 0

    pv_sum = {
        'tee:iffy': 0,
        'tee:mark': 0
    }

    async with AsyncReader(f'{pref}iffy',
                           f'{pref}mark',
                           f'{pref}counter1',
                           f'{pref}counter2',
                           mode="hasty",
                           dwell=0.0) as reader:
        while (time.time()-t0) < 5.0:
            d = await reader.wait_for_incoming()
            print('received:', d)
            received_sum += len(d)
            loop_sum += 1
            for k in pv_sum.keys():
                if k in d:
                    pv_sum[k] += 1

        # need to check that we get half as many "iffy"
        # as we get "mark", because "mark" gets updated twice
        # per IOC loop.
        assert pv_sum['tee:iffy']*2 == pv_sum['tee:mark']

        # we should receive roughly one PV at a time
        # (except for the beginning)
        assert (received_sum - loop_sum) <= 4


async def test_reader_guided(ioc):
    pref = ioc['prefix']
    t0 = time.time()
    data = {}

    received_sum = 0
    loop_sum = 0

    pvlist = (
        f'{pref}iffy',
        f'{pref}jiffy',
        f'{pref}counter1',
        f'{pref}counter2'
    )

    async with AsyncReader(*pvlist,
                           mode="guided",
                           guide_dict={ f'{pref}mark': 0 },
                           dwell=0.0) as reader:
        while (time.time()-t0) < 5.0:
            d = await reader.wait_for_incoming()
            print('received:', [k for k in d.keys()])
            for pv in pvlist:
                ## the first few data sets might not be complete with dwell == 0.0
                if (time.time()-t0) > 1.0:
                    assert pv in d


async def test_reader_thorough(ioc):
    pref = ioc['prefix']
    t0 = time.time()
    data = {}

    received_sum = 0
    loop_sum = 0

    pvlist_ok = [
        f'{pref}iffy',
        f'{pref}jiffy',
        f'{pref}mark',
    ]

    # This one will eventually block, because counterX are not being updated
    pvlist_broken = pvlist_ok + [
        f'{pref}counter1',
        f'{pref}counter2'
    ]

    async with AsyncReader(*pvlist_ok, mode="thorough", dwell=0.0) as reader:
        while (time.time()-t0) < 5.0:
            d = await reader.wait_for_incoming()
            print('received:', [k for k in d.keys()])
            for pv in pvlist_ok:
                assert pv in d

    #async with AsyncReader(*pvlist_broken, mode="thorough", dwell=0.0) as reader:
    #    #with pytest.raises
    #    d = await reader.wait_for_incoming(timeout=3.0)
                


async def test_chain_sloppy(ioc):

    src = EpicsSource(iffy=ioc['prefix']+"iffy",
                      jiffy=ioc['prefix']+"jiffy",
                      mark=ioc['prefix']+"mark",
                      mode='sloppy',
                      dwell=0.2)

    tmax = 5.0
    t0 = time.time()
    elapsed = 0

    #async with PipeExecutor(src) as ex:
    try:
        await src.startup()
        while elapsed < tmax:
            try:
                data, tmp = await asyncio.gather(src(), asyncio.sleep(1.0),
                                                 return_exceptions=False)
                print(f'Received: {[k for k in data]}')
                
                elapsed = time.time()-t0

                assert len(data) == 3
                    

            except Exception as e:
                traceback.print_exc()
                await asyncio.sleep(1.0)
                raise

    finally:
        await src.shutdown()


async def test_chain_guided(ioc):

    pref = ioc['prefix']
    src = EpicsSource(iffy=ioc['prefix']+"iffy",
                      jiffy=ioc['prefix']+"jiffy",
                      guide=f'{pref}mark==0',
                      dwell=0.2)

    tmax = 5.0
    t0 = time.time()
    elapsed = 0

    try:
        await src.startup()
        while elapsed < tmax:
            try:
                data, tmp = await asyncio.gather(src(), asyncio.sleep(1.0),
                                                 return_exceptions=False)
                print(f'Received (guided): {[k for k in data]}')
                
                elapsed = time.time()-t0

                ## tee:mark is going to be missing in guided mode
                assert len(data) == 2
                    

            except Exception as e:
                traceback.print_exc()
                await asyncio.sleep(1.0)
                raise

    finally:
        await src.shutdown()




from camagick.processor import ProcessorBase

import logging, time, asyncio, re
import numpy as np

logger = logging.getLogger(__name__)

class _FilterBase(ProcessorBase):
    # We're going to use this for both whitelist and blacklist
    # based filtering, but the blacklist one is named 'except'
    # and located elsewhere.
    
    def __init__(self, allow, *items):
        self._list = [ re.compile(i) for i in items ]
        self._allow = allow

    async def startup(self):
        pass

    async def shutdown(self):
        pass

    async def __call__(self, data=None, context=None):
        matcher = lambda x: np.array([
            ((l.match(x[0]) is not None) == self._allow) for l in self._list
        ]).any()
        
        return { k:v for k,v in filter(matcher, data.items()) }


class Processor(_FilterBase):
    '''
    Whitelist-based data blocker -- allows only named data to pass.

    Allows only data items to pass with keys that match any
    of the regex expressions. The match is case-sensitive.
    '''
    def __init__(self, *items: list):
        super().__init__(True, *items)
    




from camagick.processor import ProcessorBase

def populate_argparser(parser):
    pass

class Processor(ProcessorBase):
    '''
    Changes the array shape of the input data.
    '''
    
    def __init__(self, *dims, shape=None, **kw):
        super().__init__(**kw)
        if len(dims)==0:
            self.shape = shape
        else:
            self.shape = dims

    async def __call__(self, data=None, context=None):
        #shape = (context.get(f'{k}_shape', self.shape) for k in data) \
        #    if context is not None else self.shape
        shape = self.shape
        
        if shape is None:
            raise RuntimeError(f'msg="{self.__class__.__name__} requires a shape '
                               f'for {k} got none"')

        return { k:v.reshape(shape) for k,v in data.items() }



#!/usr/bin/python3

from numpy import zeros, sqrt
import math, time

class FpsCounter:
    def __init__(self, pts=100, label="FPS"):
        self.told = time.time()
        self.told_diff = 0
        self.fps_history = zeros(pts)
        self.ji = 0
        self.label = label
        
    def mark(self):
        tnow = time.time()
        tdiff = 1/(tnow-self.told)
        self.fps_history[self.ji] = tdiff
        self.ji = (self.ji+1)%self.fps_history.shape[0]
        self.told = tnow

    @property
    def fps(self):
        return self.fps_history.mean()

    @property
    def jitter(self):
        return sqrt( ((self.fps_history-self.fps_history.mean())**2).mean() )

    def __str__(self):
        return "%s: %2.1f ± %d Hz" % (self.label, self.fps, math.ceil(self.jitter))


    def __repr__(self):
        return str(self)



#!/usr/bin/python3

from matplotlib import pyplot as plt
from matplotlib import transforms
from epics import caget

import numpy as np

import os, sys, time, logging, math

import lmfit.models as fitmod
from lmfit import create_params

def model_fit(data, modelName, **params):    
    
    Model = getattr(fitmod, modelName)
    model = Model()
    x = np.linspace(0, len(data)-1, num=len(data))

    if len(params) > 0:
        par = create_params(**params)
    else:
        par = model.guess(data=data, x=x)

    return model.fit(data, par, x=x)


if __name__ == "__main__":

    prefix = sys.argv[1] if len(sys.argv)>1 else "KMC3:XPP:GONIOCAM:"

    bgcolor='#404050'
    fgcolor='w'
    
    fig = plt.figure(facecolor=bgcolor)
    grid = fig.add_gridspec(2, 2, width_ratios=[4,1], height_ratios=[4,1])
    ax_pic = fig.add_subplot(grid[0,0], facecolor=bgcolor)
    ax_xsum = fig.add_subplot(grid[0,1], facecolor=bgcolor)
    ax_ysum = fig.add_subplot(grid[1,0], facecolor=bgcolor)
    ax_xsum.set(xticks=[])
    ax_xsum.set_mouseover(True)
    ax_ysum.set(yticks=[])
    ax_ysum.set_mouseover(True)
    grid.tight_layout(fig)
    
    obj_pic = ax_pic.imshow( (np.array([[0, 1], [1, 0]])*256).astype(np.uint64), aspect='equal')
    obj_xsum = ax_xsum.plot([])[0]
    obj_ysum = ax_ysum.plot([])[0]
    obj_xfit = ax_xsum.plot([], linestyle='dotted')[0]
    obj_yfit = ax_ysum.plot([], linestyle='dotted')[0]
    

    obj_xstat = ax_xsum.plot([], marker='+', linestyle='solid')[0]
    obj_ystat = ax_ysum.plot([], marker='+', linestyle='solid')[0]

    fig.show()
                
    frm_shape = caget(prefix+"frameShape").astype(int)
    print("Shape:", frm_shape)
    
    obj_pic.set_extent((0, frm_shape[1], 0, frm_shape[0]))

    axis_x = np.array(range(frm_shape[0]))
    axis_y = np.array(range(frm_shape[1]))
   
    ax_xsum.set_ylim(0, frm_shape[0])
    ax_ysum.set_xlim(0, frm_shape[1])

    ax_xsum.set_xlim(0, 1)

    ax_ysum.set_ylim(0, 1)
            
    
    while True:
        try:
            frm_flat = np.array(caget(prefix+"flatFrame"), dtype=int).reshape(frm_shape)
            
            frm_x = np.array(caget(prefix+"frameSumX"), dtype=int)
            frm_y = np.array(caget(prefix+"frameSumY"), dtype=int)

            try:
                if len(frm_x) and len(frm_y):
                    fit_x = model_fit(frm_x, "GaussianModel")
                    ff_x = fit_x.eval()
                    
                    fit_y = model_fit(frm_y, "GaussianModel")
                    ff_y = fit_y.eval()

                    frm_com = np.array([f.params['center'].value for f in (fit_x, fit_y)])
                    frm_std = np.array([f.params['fwhm'].value for f in (fit_x, fit_y)])
                    #print ("fit_x", fit_x.params)

                    print("\rFWHM:"
                          " %.0f×%.0f μm, " % tuple([f.params['fwhm'].value * 3.45 for f in (fit_x, fit_y)]),
                          "σ: %.0f×%.0f μm" % tuple([f.params['sigma'].value * 3.45 for f in (fit_x, fit_y)]),
                          "                            ",
                          end='')
            except Exception as ee:
                print("No fit:", str(ee))
            
            #frm_com = caget(prefix+"centerOfMass")
            #frm_std = caget(prefix+"standardWidth")
            #print("com:", frm_com, "std:", frm_std)
            
            frame = np.array(frm_flat)
            
            obj_pic.set_data(frame[::-1,:])
            
            obj_xsum.set_data(frm_x/frm_x.max(), axis_x)
            obj_ysum.set_data(axis_y, frm_y/frm_y.max())

            obj_xfit.set_data(ff_x/ff_x.max(), axis_x)
            obj_yfit.set_data(axis_y, ff_y/ff_y.max())

            
            stdf = 0.5#math.sqrt(2*math.log(2))
            
            obj_ystat.set_data([frm_com[1]-stdf*frm_std[1],
                                frm_com[1],
                                frm_com[1]+stdf*frm_std[1]],
                               [0.5, 0.5, 0.5])
            
            obj_xstat.set_data([0.5, 0.5, 0.5],
                               [frm_com[0]-stdf*frm_std[0],
                                frm_com[0],
                                frm_com[0]+stdf*frm_std[0]])
            
        except Exception as e:
            logging.error(e)
            raise
    
        fig.canvas.draw_idle()
        fig.canvas.flush_events()
        time.sleep(0.01)



from camagick.pipe.npfun import Processor as S1Processor

from numpy.random import rand as rand_array

async def test_npstat1():

    func = [
        "mean",
        "average",
        "sum",
        "std"
    ]

    inp = rand_array(17, 43, 59)

    for f in func:
        r1 = await S1Processor(f)({'data': inp})
        for n,d in r1.items():
            print(n, d, d.shape, inp.shape)



from camagick.processor import ProcessorBase
import numpy as np


class Processor(ProcessorBase):    
    '''
    Introduces a new axis (time) and stacks data along it.

    The processor keeps a copy of all historical data of
    a particular tag, up to `size` (see initialization) number
    of data points. It acts as a ring buffer, i.e. once the
    maximum number of points is reached, new older points get
    discarded as newer ones are added.
    '''

    def __init__(self, size: int = 128, reslice: bool = True):
        '''
        Args:
            size: number of points to keep stacked. Older points
              get discarded as newer ones are added.

            reslice: scalar values and 1D arrays with a single element
              are sometimes used interchangably in Caspy. Usually this
              isn't a problem, but here it becomes one: stacking scalars
              results in a 1D array (which can be easily plotted), while
              stacking a 1D / single element array results in a 2D array
              (with the 2nd dimension of size 1), which *cannot* be plotted
              and must be either sliced, or treated as an image.

              Setting this to True is a convenience option to actually
              do the reslicing on the fly in such cases.

              Default is True.
              
        '''

        self._ring_size = size
        self._stacks = {}
        self._reslice = reslice
        
    
    async def __call__(self, data=None, context=None):

        if data is None:
            return None

        for tag, inp in data.items():
            stack_shape = (self._ring_size, *inp.shape)
            try:
                prev = self._stacks[tag]
            except KeyError:
                prev = np.ndarray(stack_shape)
                prev[...] = inp.mean()

            cur = self._stacks[tag] = np.roll(prev, -1, 0)
            cur[-1] = inp

        output = {}

        for k in data:
            stk = self._stacks[k]
            if len(stk.shape) == 2 and stk.shape[1] == 1 and self._reslice:
                output[k] = stk[:,0]
            else:
                output[k] = stk
            
        return output



#!/usr/bin/python3

from matplotlib import pyplot as plt
from matplotlib import transforms

import numpy as np
import xarray
import asyncio
import os, sys, time, logging, math

import argparse

from camagick.helpers import FpsCounter
from camagick.stash import ZarrStash

class ImageMplWidget:
    ''' Matplotlib-based image display widget '''

    def __init__(self):
        bgcolor='#404050'
        fgcolor='w'
    
        self.fig = plt.figure(facecolor=bgcolor)
        self.grid = self.fig.add_gridspec(2, 2, width_ratios=[4,1], height_ratios=[4,1])
        
        self.ax_pic = self.fig.add_subplot(self.grid[0,0], facecolor=bgcolor)
        self.obj_pic = self.ax_pic.imshow( (np.array([[0, 1], [1, 0]])*256).astype(np.uint64), aspect='auto')
        
        #ax_xsum = fig.add_subplot(grid[0,1], facecolor=bgcolor)
        #ax_ysum = fig.add_subplot(grid[1,0], facecolor=bgcolor)
        #ax_xsum.set(xticks=[])
        #ax_xsum.set_mouseover(True)
        #ax_ysum.set(yticks=[])
        #ax_ysum.set_mouseover(True)
        self.grid.tight_layout(self.fig)
    
        #obj_xsum = ax_xsum.plot([])[0]
        #obj_ysum = ax_ysum.plot([])[0]

        #obj_xstat = ax_xsum.plot([], marker='+', linestyle='solid')[0]
        #obj_ystat = ax_ysum.plot([], marker='+', linestyle='solid')[0]

        self.fig.show()
                
        #frm_shape = caget(prefix+"frameShape").astype(int)
        #print("Shape:", frm_shape)
    
        #obj_pic.set_extent((0, frm_shape[1], 0, frm_shape[0]))
        
        #axis_x = np.array(range(frm_shape[0]))
        #axis_y = np.array(range(frm_shape[1]))
        
        #ax_xsum.set_ylim(0, frm_shape[0])
        #ax_ysum.set_xlim(0, frm_shape[1])
        
        #ax_xsum.set_xlim(0, 1)
        #ax_ysum.set_ylim(0, 1)


    def update(self, image):
        #frm_flat = np.array(caget(prefix+"flatFrame"), dtype=int).reshape(frm_shape)

        #frm_x = np.array(caget(prefix+"frameSumX"), dtype=int)
        #frm_y = np.array(caget(prefix+"frameSumY"), dtype=int)

        #frm_com = caget(prefix+"centerOfMass")
        #frm_std = caget(prefix+"standardWidth")

        #frame = np.array(frm_flat)

        img = image[::-1,:]
        self.obj_pic.set_data(img)
        #self.obj_pic.set_extent((0, img.shape[0], 0, img.shape[1]))
        
        #obj_xsum.set_data(frm_x/frm_x.max(), axis_x)
        #obj_ysum.set_data(axis_y, frm_y/frm_y.max())


        #stdf = math.sqrt(2*math.log(2))

        #obj_ystat.set_data([frm_com[1]-stdf*frm_std[1],
        #                    frm_com[1],
        #                    frm_com[1]+stdf*frm_std[1]],
        #                   [0.5, 0.5, 0.5])

        #obj_xstat.set_data([0.5, 0.5, 0.5],
        #                   [frm_com[0]-stdf*frm_std[0],
        #                    frm_com[0],
        #                    frm_com[0]+stdf*frm_std[0]])

        self.fig.canvas.draw_idle()
        self.fig.canvas.flush_events()


def get_options():
    parser = argparse.ArgumentParser(prog='camagick-display-image',
                                     description='Display a 2D data array using imshow()')

    parser.add_argument('--zarr', action='store',
                        help='ZARR file to watch for changes')
    parser.add_argument('--lock', action='store',
                        help='Lock key to use for synchronising access to data')

    parser.add_argument('-cdf', action='store',
                        help='netCDF4 file to watch for changes')

    return parser.parse_args()


def image_main():
    asyncio.run(async_main())

    
async def async_main():
    args = get_options()
    image_plot = ImageMplWidget()

    stash = ZarrStash(args.zarr or args.cdf, lockKey=args.lock)
        
    fps = FpsCounter(100)

    while True:
        try:
            ds = stash.load()
            image_plot.update(ds['flatFrame'])            
            
        except Exception as e:
            logging.error("Stash retrieve failed: %r", e)

        fps.mark()
        print("\r%r      " % fps, end='')    


if __name__ == "__main__":
    async_main()



#!/usr/bin/python3

import logging, asyncio, importlib, traceback

from camagick.processor import ProcessorBase, SourceBase
from camagick.probe import opcast

from ophyd.utils import errors as ophyd_errors
from ophyd import Device, Component
from ophyd.signal import EpicsSignalRO

import numpy as np

import caproto

logger = logging.getLogger(__name__)

class TestDevice(Device):
    # This is just a rapid test object to learn how to
    # deal with multi-component Ophyd devices, what makes
    # them tick, and what do we need to support in "Processor".
    #
    # For this to work, run a test IOC, e.g. like this:
    #
    #   caspy --from-generate random jiffy=100 miffy=100 --to-ioc prefix=foo:
    #
    # Then you can test Ophyd device readout like this:
    #
    #   caspy --from-ophyd camagick.source.ophyd:TestDevice name=foo prefix=foo: --to-plot
    #
    
    miffy = Component(EpicsSignalRO, "miffy")
    jiffy = Component(EpicsSignalRO, "jiffy")

class Processor(SourceBase):
    '''
    Retrieves data from an Ophyd device.

    Examples
    --------

    - Classig `EpicsSignalRO` readout:
      ```
      caspy --from-ophyd ophyd.signal:EpicsSignalRO EXAMPLE:iffy ...
      ```
    '''

    def __init__(
            self,
            cls: str,
            *oa: list,
            mode: str = None,
            opcast: bool = True,
            trigger: bool = False,
            stage: str = "never",
            **ookw: dict
    ):
        '''
        Initializes an Ophyd device from a user-specified module.

        It retrieves all the data that the device's `.read()`
        function returns, using the Ophyd keys as CAspy internal
        keys.

        Args:
            cls: Ophyd class to use, in the notation
              `module.sub:Class`

            *oa: positional arguments to pass on to
              the Ophyd class constructor

            mode: Data acquisition mode, one of "subscribe" or "poll".
              "subscribe" is the default.

            **okw: keyword arguments to pass on to the
              Ophyd class
        '''

        cls_obj = self._get_class(cls)
        self._ovar = cls_obj(*oa, **ookw)

        self._auto_opcast = opcast

        if mode is None:
            # We don't support subscribe mode for Ophyd devices
            # that don't have a "value" type event to subscribe to.
            mode = 'subscribe' if "value" in self._ovar.event_types \
                else 'poll'

        if mode == 'subscribe':
            for et in ("value", "acq_done"):
                if et  in self._ovar.event_types:
                    self._ovar.subscribe(self._data_incoming, et)
                    logger.info(f'msg="Subscribing" sub_type={et} ophyd={self._ovar.name}')
                    break
            self._retr_data = self._retr_data_subscribe
            
        elif mode == 'poll':
            self._retr_data = self._retr_data_poll
            
        else:
            logger.error(f'msg="Unsupported mode" mode={mode}')
            raise RuntimeError(f'Bad mode "{mode}"')

        self._trigger_init = False
        self._trigger_readout = trigger
        self._stage_init = stage in ("once", )
        self._stage_readout = stage in ("always", "yes")

        self._disconn_reported = None
        self._data_lock = asyncio.Lock()

        self._data_cache = {}

        super().__init__()


    def _get_class(self, cspec):
        parts = cspec.split(':')

        if len(parts) != 2:
            raise RuntimeError(f'msg="Ophyd source needs a `<module>:<class>` spec"')
        
        logger.info(f'msg="Finding Ophyd device" '
                    f'class_name="{parts[1]}" '
                    f'module_name="{parts[0]}"')
        mod_obj = importlib.import_module(parts[0])
        cls_obj = getattr(mod_obj, parts[1])

        return cls_obj


    def _name2tag(self, name):
        # Ophyd names sometimes have ':' and '.' in them.
        # We need to strip those away.
        return ''.join([
            (c if c.isalnum() else '_') for c in name
        ])


    async def _maybe_prepare(self, proc, condition):
        if hasattr(self._ovar, proc) and condition:
            await asyncio.get_event_loop().\
                run_in_executor(None, getattr(self._ovar, proc))


    async def startup(self):
        await self._data_lock.acquire()
        await asyncio.get_event_loop().run_in_executor(
            None, self._ovar.wait_for_connection
        )
        await self._maybe_prepare("stage", self._stage_init)


    async def shutdown(self):
        await self._maybe_prepare("unstage", self._stage_init)
        await asyncio.get_event_loop().run_in_executor(
            None, self._ovar.disconnect
        )        


    async def _retr_data_poll(self):
        try:
            await self._maybe_prepare("stage", self._stage_readout)
            await self._maybe_prepare("trigger", self._trigger_readout)

            tree = await asyncio.get_event_loop().\
                run_in_executor(None, self._ovar.read)

            return {
                self._name2tag(k):d['value'] for k,d in tree.items()
            }

        finally:
            await self._maybe_prepare("unstage", self._stage_readout)            


    async def _retr_data_subscribe(self):

        # There's a bug on some versions of asyncio (?) which will prevent
        # the coroutine from waking up once the lock is released. Doing this
        # in a loop apparently works around this.
        while True:
            try:
                await asyncio.wait_for(self._data_lock.acquire(), timeout=0.01)
                break                
            except TimeoutError:
                continue
            
        if isinstance(self._data_cache, dict):
            return self._data_cache
        else:
            return await self._retr_data_poll()


    def _data_incoming(self, **kw):
        etype = kw['sub_type']
        
        if etype == 'value':
            val = kw['value']
            oldv = kw['old_value']
            ts = kw['timestamp']
            obj = kw['obj']

            if self._data_cache is None:
                raise RuntimeError(f'Looks like you\'re mixing "value" and '
                                   f'"acq_done" events; don\'t do that.')

            if self._auto_opcast:
                v = opcast(val, type_sequence=(int, float, np.array, str))
            else:
                v = val
            self._data_cache[self._name2tag(obj.name)] = np.array(v)

            try: self._data_lock.release()
            except RuntimeError: pass
                
        elif etype == 'acq_done':
            # Need to retrieve data...?
            # We completely invalidate the data cache, and let _retr_data_subscribe()
            # retrieve the data by issuing one .read() opeartion.
            self._data_cache = None
            if self._data_lock.locked():
                self._data_lock.release()

        else:
            logger.error(f'msg="Unexpected event" type="{etype}"')


    async def __call__(self, data=None, context=None):
        try:
            d = await self._retr_data()
            
            if self._disconn_reported == True:
                logger.info(f'msg="Ophyd device available again" '
                            f'name="{self._ovar.name}"')
                self._disconn_reported = False

            return d

        except (ophyd_errors.DisconnectedError,
                caproto.CaprotoTimeoutError):
            if self._disconn_reported in (None, False):
                logger.warning(f'msg="Ophyd device not connected" '
                               f'name="{self._ovar.name}"')
                self._disconn_reported = True
            return {}
        
        except Exception as e:
            traceback.print_exc()
            logger.error(f'msg="Error retrieving Ophyd data" reason="{e}"')
            return None        



from camagick.processor import ProcessorBase, SinkBase

import sys

class Processor(SinkBase):
    '''
    Outputs a human-readable data summary.

    Generally, the intention is to print the data itself, as it's the
    most self-explanatory form -- but only with scalar data :-)
    For complex data types (e.g. arrays), we display a short message
    notifying _that_ it's an array, and the shape and data type it has.

    We call this "the information".
    '''
    
    def __init__(self, mode: str = 'block', fmt: str = '{tag}: {}', path: str = '-'):
        '''
        Args:
            mode: output mode, choose between 'block' or 'table'.
              In `block` mode, essentially every line is one data set,
              i.e. the complete data payload of one pipeline frame is
              displayed as a "block". The format of each block line is
              specified by `fmt`.

              In `table` mode, the keys are printed first as a header of
              sorts, and the data information is printed per-round as
              a single tab-separated line, one line per frame.

            fmt: Format of a block line, default is `{tag}: {}`, where
              `{}` is substituted by the data representation (mostly
              it's a value if the value is compact; for arrays, it's
              the shape of the array and the data type).

            path: where to write the data. If empty string '', or '-'
              data is written to `stdout`. Otherwise writing to the specified
              file is attempted.
        '''

        self._mode = mode
        self._blk_fmt = fmt
        self._path = path
        self._header_list = []

        self._current_file = None


    def value_repr(self, v):
        if hasattr(v, "shape"):
            if (len(v.shape)>1) or \
               (len(v.shape)==1 and v.shape[0]>1):
                dt = type(v.flatten()[0].item())
                return f'shape={v.shape} {dt}'
            else:
                return str(v)
                
        else:
            return f'{v}'
        

    def ensure_header(self, fobj, data):
        for h in self._header_list:
            if h not in data:
                self._header_list = [k for k in data]
                self.ensure_header(fobj, data)
                fobj.write('#' + '\t'.join([k for k in data])+'\n')

        for k in data:
            if k not in self._header_list:
                self._header_list = [k for k in data]
                self.ensure_header(fobj, data)
                fobj.write('# ' + '\t'.join([k for k in data])+'\n')

        return None


    def write_data(self, fobj, data, context):
        if self._mode in ('block'):
            for k,v in data.items():            
                fobj.write(self._blk_fmt.format(self.value_repr(v),
                                                tag=k, **context)+'\n')
            return
        
        if self._mode in ('table'):
            self.ensure_header(fobj, data)
            fobj.write('\t'.join([
                self.value_repr(data[k]) for k in self._header_list
            ])+'\n')
            return

        raise RuntimeError(f'msg="Unknown summary mode" mode={self._mode}')


    def reset_output(self):
        self._header_list = []


    def ensure_folder_of(self, path):
        d = os.path.dirname(path)
        if not os.path.exists(d):
            os.makedirs(d)
            return
        if not os.path.isdir(d):
            raise RuntimeError(f'msg="Path is required to be a folder folder" path="{p}"')        
        
    
    async def __call__(self, data=None, context=None):
        if data is None:
            return

        if self._path in ('', '-'):
            self.write_data(sys.stdout, data, context)

        else:

            p = self._path.format(**context)
            if self._current_file != p:
                self._current_file = p
                self.reset_output()

            self.ensure_folder_of(p)

            with open(p, 'ta+') as fobj:
                self.write_data(fobj, data, context)



from camagick.source.astor import Processor as ProcessorBase
import zarr

def open_data_file(path):
    return zarr.open_group(path, 'r')

class Processor(ProcessorBase):
    '''
    Short for `--astor kind=zarr`.
    '''
    def __init__(self, url: str, **data):
        super().__init__(url, kind='zarr', **data)



'''
Array Storage API abstraction

This is meant to implement the operations that CAmagick supports in
an backend-agnostic way. Typically, CAmagick writes to HDF5, Zarr
and Tiled. So this API intents to support the common denominator of
all of these:

- Eternal / internal path (i.e. an entry point, "file" or "url"
  that designates a high-level resource, and an internal path
  that designates a low-level resource within the high-level
  one)

- Groups or Datasets (HDF5-like hiearchical structure, where
  there are datasets or groups, with the groups containing further
  datasets or subgroups, accessible by ["...path..."])

- Datasets can be sliced, and meta information (in particular data type
  and shape) can be accessed by `.dtype`, respectively `.shape`

- Datasets and/or groups can have attributes by means of `.attrs[...]`

- Files can be opened for reading, or for read-writing, optionally
  overwriting or refusing to create new files when opening for
  read-writing and the file already exists / doesn't exist.
'''

import time, logging, numpy, functools, asyncio, importlib

logger = logging.getLogger(__name__)

__all__ = [
    "RefuseIndex",
    "RefuseOverwrite",
    "RefuseSparse",
    "ArrayAccess",
    "MemoryArrayAccess"
]

# list of procedures that return True if a specific storage
# path/url is of a specific format.

_format_probe = {
    'hdf5': lambda x: numpy.array([
        x.endswith(s) for s in \
        ('.h5', '.hdf', '.hdf5', '.nx', '.cdf4')
    ]).any(),
    
    'zarr': lambda x: numpy.array([
        x.endswith(s) for s in \
        ('.z', '.zarr')
    ]).any(),
}

def probe_kind(url):
    for kind, stest in _format_probe.items():
        if stest(url):
            return kind
    raise RuntimeError(f'no matching loader found for {url}')

def find_loader(url=None, kind=None):
    if kind is None:
        kind = probe_kind(url)
    mod = importlib.import_module(f'camagick.storage.{kind}')
    return getattr(mod, 'ArrayAccess')


class RefuseIndex(RuntimeError): pass

class RefuseOverwrite(RefuseIndex): pass

class RefuseSparse(RefuseIndex): pass

class DatasetUnavailable(RuntimeError):
    def __init__(self, tag, reason=None):
        self.tag = tag
        super().__init__(f'Dataset unavailable: "{tag}" ({reason})')

class ArrayAccess:
    '''
    This is just to document the abstract API.

    The API elements are heavily tailored towards CAmagick's needs, which
    are essentially appending new data ("frames") to existing datasets.
    To implement a new backend, at least the following members need
    to be defined:

    - Store opening: `_open_writable()`
    - Dataset management: `_ensure_dataset_base()`, `_ensure_dataset()`

    Additonally, possibly:
    - Read-only access: `_open_readable()` if different from read-write access
    - Store closing: `_close()`, if the store object doesn't support `.close()`    
    - Dataset sizing: `._resize_dataset()` (and possibly `_reset_dataset()`),
      if the dataset doesn't have a `.resize(shape)` function.
    '''

    def __init__(self, store=None, sdir=None, mode=None, **tags):
        '''
        Initializes the storage mechanism.

        This is typically created by higher layers, once per session (i.e.
        read/write loop run). Saving to multiple locations (i.e. store "files"
        paths within files) is supported by virtue of appropriate combination
        of `sdir` and/or `**tags` parameters. The actual storage backend objects
        are created lazily, during async enter/exit environments. This makes it
        possible to use formatting spec `{...}` in the storage paths, and
        set those paths at a later time using format keys.
        
        
        Args:
            store: Top-level store path (file name or URL). May contain
              formatting values `{...}`, which are evaluated later at
              `.push_frame()` time.

            sdir: Path inside the store. If this is not `None`, it is
              prepended to any of the tags logation from `tags`. It is
              intended to make it easy to specify a top-level directory,
              or to make a one-size-fits-all path spec for all tags.
              May contain formatting values `{...}`, which are evaluated
              at `.push_frame()` time.

            mode: What to do when we're required to create or overwrite a
              data point. We only concert ourselves with data points since
              this is what CAmagick's basic data currency is. For everything
              else (files, groups), the policy follows whatever must be done
              to ensure the data point mode policy can be followed.
        
              This must be either `None`, or a string consistinc or a
              combination of `a` (append), `o` or `O` (overwrite)
              or `x` (reset), possibly in combination with `+` (top-up),
              meaning:
        
                 - `a`: require that the data point be always appended
                   to the current dataset -- never overwrite or delete
                   existing data.
        
                 - `o`: overwrite already existing data points (frames).
                   If the shape of the new data point frames doesn't match
                   the old ones, then fail.

                 - `O`: same as `o`, but additioally delete and re-create
                   whole dataset entries, if the shape or data type of the
                   existing dataset does not match the new request.

                 - `x`: reset the entire dataset (i.e. delete all existing
                   data points) if a data point already exists at the
                   specified index.

                 - `+`: top-up marker will fill up empty slots, but without
                    actually writing any data to them. They will remain
                    filled with default settings of the corresponding backend
                    and the data type. If top-up is not specified, otherwise
                    the operation will fail if index is higher than the one of
                    the next slot in turn.

              If `mode` is set to `None`, each dataset, and consequently the
              whole store, is opened for reading only. This also reflects
              the behavior of the async environment enter/exit functionality.

            **tags: Keys are tag names (by which to refer the data in
              `.push_frame()`, and values are path elements, for each
              frame in itself. If the values are `None`, the value of `sdir`
              is used. Otherwise the value of `sdir` (if itself not `None`)
              is prepended to the path values.
        '''
        self._stor_spec = (store, sdir)
        self._stor_dspath = {
            k:(sdir if sdir is not None else '')+(v if v is not None else '') \
            for k,v in tags.items()
        }
        self._stor_timeout = 1.0
        self._refcnt = 0
        self._frame_mode = mode
        #self._stor_keys = {}

    
    #self._stor_keys = kw.copy()


    @property
    def readonly(self):
        return self._frame_mode is None


    async def __aenter__(self):
        self._refcnt += 1        
        if self._refcnt > 1:
            return self

        if self._frame_mode is None:
            self._stor_open = functools.partial(self._open_readable)
        else:
            self._stor_open = functools.partial(self._open_writable)


        # Typically we are accessing several _arrays_ from one _file_
        # (or from a smaller number of files). We can't close files
        # or generally file-like handles (sockets) as long as there
        # are still arrays using it. So we need to implement a
        # file-open caching refcounter.
        # This holds the final store path (i.e. "file path") as known
        # to us as key, and the store refcounter (see ._open()) as value.
        # So this is where first part of the magic happens: we store
        # a StorObj for every data tag we want to save. We don't
        # have any of the tags available here, but we initialize
        # the cache and store the keys.

        self._stor_fileobj = {}
        objlist = await asyncio.gather(
            *[self.__open_store(k, timeout=self._stor_timeout) \
              for k in self._stor_dspath],
            return_exceptions=False
        )

        self._stor_obj = {
            k:v for k,v in zip(self._stor_dspath, objlist)
        }

        return self


    async def __open_store(self, tag, timeout=1.0):
        # Generic (backend-agnostic) store open(). This is calle per _tag_,
        # i.e. possibly multiple times per storage backend. We have a pretty
        # confusing refcount system to ensure that every backend is opened
        # only once, and closed when the last dataset is being discarded.

        class StorRefcnt:
            ## Store base object ("file")
            fo = None

            ## When this reaches 0 again, we can close .fo
            refcnt = 0
            

        class StorObj:
            ## Refcounter object for the store "file" object
            fref = None

            ## Base object for dataset (whether this is
            ## "file" or "group", backend specific)
            bo = None

            ## Dataset path inside the storage
            dpath = None
        
        obj = StorObj()
        stor = self._stor_spec[0]
        path = (self._stor_spec[1] if self._stor_spec[1] is not None else '') + \
            self._stor_dspath.get(tag, '')

        # Some storage backends have locks which make opening operations
        # fail temporarily. Try several times until a final timeout.
        # The "file" here will stay open and needs to be closed with
        # `._close(tag)`. At API level, use aenter/aexit!

        # Find the storage backend first
        try:
            # Store backend already open -- reuse
            fref = self._stor_fileobj[stor]
            fref.refcnt += 1

        except KeyError:
            # Store backend needs to be opened first
            fref = self._stor_fileobj[stor] = StorRefcnt()
            fref.refcnt = 1
            t0 = time.time()
            while True:
                try:
                    fref.fo = await self._stor_open(stor)
                    break
                except Exception as e:
                    if (time.time()-t0) >= timeout:
                        raise e
                    await asyncio.sleep(1e-3)

        try:
            obj.fref = fref
            obj.bo, obj.dpath = self._ensure_dataset_base(obj.fref.fo, path)
            return obj
        except Exception as e:
            err = f'msg="Cannot ensure base of dataset" dset={tag} reason="{e}"'
            logger.error(err)
            self.__close_store(obj=obj)
            raise DatasetUnavailable(tag, err)


    async def __aexit__(self, *args):
        self._refcnt -= 1
        if self._refcnt < 0:
            raise RuntimeError('you are not supposed to be here')
        
        for k,o in self._stor_obj.items():
            self.__close_store(obj=o)
        self._stor_obj = None


    async def _open_readable(self, store):
        await self._open_writable(store) # default implementation


    async def _open_writable(self, store):
        raise RuntimeError('not implemented')


    def _close(self, stor_obj):
        '''
        Closes the store object referenced by `stor_obj`.
        The default implementation just calls a `.close()` if it finds one,
        because that's what most file or socket based backends would do.
        You may need to reimpement this.
        '''
        stor_obj.close()
        

    def __close_store(self, tag=None, obj=None):
        # FIXME: check that every file is closed only _once_,
        # even if there are several tags that point to the
        # same file!
        if obj is None:
            obj = self._stor_obj[tag]

        if obj.fref.refcnt <= 0:
            # This is a bug
            raise RuntimeError(f'double-close for "{tag}"?')

        obj.fref.refcnt -= 1

        if obj.fref.refcnt == 0:
            self._close(obj.fref.fo)

            
    def _split_dset_path(self, dset_path, splitter='/'):
        # Helper that will return the (node path, dataset path)
        # out of a whole path.
        # If there's only one component, assume it's a dataset.
        parts = dset_path.split(splitter)
        
        if len(parts) == 0:
            raise RuntimeError(f'invalid dataset path {dset_path}')

        if len(parts) == 1:
            return ('', parts[0])

        return ('/'.join(parts[:-1]), parts[-1])
               


    def _ensure_dataset_base(self, stor, dset, **attrs):
        '''
        Makse sure the base for storing the dataset exists.

        Args:
            stor: The storage object (file or group)
        
            dset: Dataset path relative to `stor`
        
            **attrs: ?

        Returns a `(baseobj, dset_path)` tuple. `dset_path` here needs
        to be relative to `baseobj`.
        '''
        raise RuntimeError('not implemented')


    def _ensure_dataset(self, path, base, dtype=None, frame_shape=None, **attrs):
        '''
        Ensure the existence of the target dataset.
        
        This is typically called with the first received frame-push (i.e.
        data point), and is intended to set sets names, types,
        attributes etc.

        Args:
            base: where to put the dataset; for some storage backends
             this is "file" object, for others this is the group object
             that immediately holds the dataset

            path: full path of the dataset inside `base`
        
            dtype: data type (numpy-dtype); if array is in read-only
              mode, this and `frame_shape` are expected to be `None`,
              and the actual data type and frame shape of the target
              object will be ignored.
        
            frame_shape: shape of the data point; note that
              the actual shape of the resulting data set will contain
              one extra dimension upfront (i.e. the frame index).
        
            *attrs: dataset attributes

        Return the dataset object.
        '''
        raise RuntimeError('not implemented')


    def _resize_dataset(self, dset, new_shape):
        '''
        Helper for `._ensure_frame()`.

        Essentially, this is the implementation-dependent part. Most backends
        will simply offer a `.resize()` call (this is true for HDF5 and Zarr),
        which is also the default implementation here.

        But others may require a more elaborate procedure (e.g. saving data,
        resizing, restoring).
        '''
        dset.resize(new_shape)


    def _reset_dataset(self, dset):
        '''
        Helper for `._ensure_frame()`.

        Erases all data points from a dataset.
        '''
        new_shape = [int(i) for i in (0, *(dset.shape[1:]))]
        self._resize_dataset(dset, new_shape)


    def _ensure_frame(self, dset, index):
        '''
        Ensures that `dset` has at least `index` points in its first dimension.

        Args:
            dset: Dataset object, dependent on the backend

            index: An integer counting the objects (starting with 1?)

        '''
        idiff = (index-dset.shape[0]) + 1

        if idiff == 1:
            shape_diff = 1
            
        elif idiff > 1:
            if '+' in self._frame_mode:
                shape_diff = idiff
            else:
                raise RefuseSparse(f'msg="Sparse writing requires top-up flag" '
                                   f'index={index} frames={dset.shape[0]} ')
            
        elif idiff <= 0:
            if 'a' in self._frame_mode: # append mode
                raise RefuseOverwrite(f'msg="Overwrite not permitted" index={index} '
                                      f'frames={dset.shape[0]} dataset="{dset.name}" '
                                      f'file=""')

            elif 'o' in self._frame_mode or 'O' in self._frame_mode: # overwrite mode
                shape_diff = 0

            elif 'x' in self._frame_mode: # reset mode
                logger.info(f'msg="Resetting dataset" index={index} '
                            f'frames={dset.shape[0]} dataset="{dset.name}"')
                self._reset_dataset(dset)

                shape_diff = index+1

            else:
                raise RuntimeError(f'BUG: No shape diff for mode: {self._frame_mode}')

        if shape_diff > 0:
            new_shape = [int(i) for i in (dset.shape[0]+shape_diff, *(dset.shape[1:]))]
            logger.debug(f'msg="Resizing dataset" '
                         f'size={dset.shape[0]} '
                         f'index={index} diff={shape_diff} idiff={idiff} '
                         f'shape={new_shape}')
            self._resize_dataset(dset, new_shape)


    async def push_frame(self, index, **data):
        '''
        Received for every data point in a scan.

        Args:
            index: Integer to determine the position of the current
              data point in the overall (larger) dataset. This should
              always point to the next available index in the first
              dimention, i.e. if the existing dataset has dimension
              (37, N, M), with the current data point to become
              the 38th, then `index` needs to be 37 (counting starts
              at 0).

              Behavior if the index doesn't match the shape is defined
              by the class-global parameter `mode`.

              If this is `None`, the next free slot (i.e. current_size+1)
              is used, essentially making the operation always succeed
              as if "append" mode were active.

              If this is a negative number (e.g. -1) and the dataset
              doesn't exist, it is created according to shape
              specification of the input data, but with length 0
              (i.e. no points are copied).
              
            **data: dictionary from data name ("tag") to data arrays
        '''

        for key,container in data.items():
            obj = self._stor_obj[key]
            dset = self._ensure_dataset(obj.dpath,
                                        obj.bo,
                                        container.dtype,
                                        container.shape)

            if index is None:
                index = container.shape[0]

            if index >= 0:
                self._ensure_frame(dset, index)
                item = numpy.array(container)
                dset[index] = item


    def __getitem__(self, tag):
        '''
        Returns a handle to the object internally known as `tag`, according
        to the path spec specified at `__init__` time. This is generally
        supposed to be a dataset.
        '''
        obj = self._stor_obj[tag]
        return self._ensure_dataset(obj.dpath, obj.bo)




from camagick.processor import ProcessorBase

import logging, time, asyncio, re
import numpy as np

logger = logging.getLogger(__name__)

class Processor(ProcessorBase):
    '''
    Allow only non-empty datasets to pass
    '''
    async def __call__(self, data=None, context=None):
        return {
            k:v for k,v in filter(lambda x: len(x[1].flatten())>0,
                                  data.items())
        }



#!/usr/bin/python3

import pytest, random, string, asyncio, os, time
from camagick import stash

@pytest.mark.asyncio
async def _test_stash_lock():
    rnd = ''.join(random.choice(string.ascii_lowercase) for i in range(24))

    st = stash.ZarrStash("/tmp/storage"+rnd, lockKey=rnd)

    #print("Waiting for lock...")
    await st.asyncWaitLock()

    pid = os.fork()

    if pid == 0:

        await st.asyncWaitLock()
        
        # parent, child is `pid`
        time.sleep(1.0)

        #print("Unlocking")
        st.unlock()
        
    else:
        # child
        #print("Try to relock (should fail)")
        assert st.tryLock() == False

        time.sleep(2.0)

        #print("Try to relock (should work)")
        assert st.tryLock() == True
        
        #print("Unlocking again")
        st.unlock()


@pytest.mark.asyncio
async def _test_stash_context():
    rnd = ''.join(random.choice(string.ascii_lowercase) for i in range(24))

    async with stash.ZarrStash("/tmp/storage"+rnd, lockKey=rnd) as st:

        pid = os.fork()
        
        if pid == 0:
            # parent
            await st.asyncWaitLock()
            time.sleep(1.0)
        
        else:
            # child
            assert st.tryLock() == False
            time.sleep(2.0)
            assert st.tryLock() == True
            st.unlock()
        
        



#!/usr/bin/python3

import logging, time, math, random, asyncio

from camagick.processor import ProcessorBase, SinkBase
from camagick.probe import opcast

import numpy as np

import tango
from tango import asyncio as tas

logger = logging.getLogger(__name__)

class Processor(SinkBase):
    '''
    Sends data to a tango device
    '''

    def __init__(self, dev: str, **tags):
        '''
        Produces data.

        Args:
            **tags: tag -> tango variable
        '''
        self._dev = dev
        self._tags = tags


    async def startup(self):
        self._proxy = await tas.DeviceProxy(self._dev)
        logger.info(f'msg="Have device proxy" proxy={self._proxy}')


    async def shutdown(self):
        pass


    async def __call__(self, data=None, context=None):

        awaitables = []
        if data is None:
            data = {}
        
        for k in data:
            attr = self._tags[k]
            val = data[k]
            logger.info(f'msg="Setting" device="{self._dev}" attr={attr} value={val}')
            awaitables.append(self._proxy.write_attribute(attr, val))

        await asyncio.gather(*awaitables)



from camagick.executor import PipeExecutor
from camagick.storage.memory import ArrayAccess as MemoryArrayAccess

import tempfile, os, random, h5py, pytest, pprint, importlib
import numpy as np

for mod in ('astor',):
    name_here = mod[0].upper()+mod[1:]
    base = importlib.import_module(f'camagick.source.{mod}')
    globals()[name_here] = getattr(base, 'Processor')

@pytest.fixture()
def rndata():
    return np.random.rand(30, 36, 72)

@pytest.fixture()
def h5tmp(rndata):
    tmp = tempfile.NamedTemporaryFile(delete=False)
    tmp.close()
    with h5py.File(tmp.name, 'a') as h5:
        grp = h5.create_group('/group')
        grp.attrs['gmeta'] = 'group'
        data = grp.create_dataset(data=rndata, name='data')
        data.attrs['dmeta'] = 'dataset'
    yield tmp.name
    #os.unlink(tmp.name)


async def test_astor(h5tmp, rndata):

    ast = Astor(
        h5tmp,
        kind='hdf5',
        full='/group/data',
        part='/group/data[1:10:2,10:30]',
        dmeta='/group/data.attrs[dmeta]',
        gmeta='/group.attrs[gmeta]',
        dshape='/group/data.shape',
        dlist='/group.keys',
        root='.keys'
    )
    data = await ast()

    assert (data['full'] == rndata).all()
    assert (data['part'] == rndata[1:10:2,10:30]).all()
    assert data['dshape'] == rndata.shape
    assert data['dmeta'] == 'dataset'
    assert data['gmeta'] == 'group'
    assert data['dlist'][0] == 'data'
    assert 'group' in data['root']


async def test_array_store():
    s1 = MemoryArrayAccess("testarray",
                           mode='a+',
                           d1="group/data1",
                           d2="group/sub/data2")

    async with s1:
        for i in range(10):
            await s1.push_frame(i,
                                d1=np.array([1*i, 2*i, 3*i]),
                                d2=np.array(3.14*i))

    pprint.pprint(s1._repo._storage)
    



from camagick.source.astor import Processor as ProcessorBase
import h5py

def open_data_file(path):
    return h5py.File(path, 'r')

class Processor(ProcessorBase):
    '''
    Short for `--astor kind=hdf5`.
    '''
    def __init__(self, url: str, **data):
        super().__init__(url, kind='hdf5', **data)



from camagick.processor import ProcessorBase, SinkBase
from camagick.ioc import InstaOutputPVDB, IocServer
import numpy as np
import logging, asyncio, traceback


logger = logging.getLogger(__name__)


class Processor(SinkBase):
    ''' Exports data as EPICS PVs '''

    def __init__(self, prefix=None, suffix=None):
        self.ioc = InstaOutputPVDB()
        self.ioc_prefix = prefix if prefix is not None else ''
        self.ioc_suffix = suffix if suffix is not None else ''
        self.ioc_server = IocServer(self.ioc)
        
        self.pv_format = f'{{prefix}}{{tag}}{{suffix}}'
        logger.info(f'msg="Sink IOC" pv_format="{self.pv_format}"')


    async def startup(self):
        await self.ioc_server.startup()


    async def shutdown(self):
        await self.ioc_server.shutdown()


    async def __call__(self, data=None, context=None):

        if data is None:
            data = {}

        if context is None:
            context = {}

        try:
            await self.ioc.publish_many({
                self.pv_format.format(prefix=self.ioc_prefix, tag=name,
                                      suffix=self.ioc_suffix, **context) : val \
                for name,val in data.items()
            })

        except:
            traceback.print_exc()

        return None



def find_applications(prefix, pattern='*', exclude=None):
    apps = []
    for p in os.environ['PATH'].split(':'):
        apps += [p+"/"+x for x in glob.glob("%s-%s" % (prefix, pattern), root_dir=p)]

    result = {}
    for a in apps:
        name = os.path.basename(a).replace(prefix+'-', '')

        if name.split('-')[0] in (exclude or []):
            continue
        
        result[name] = a

    return result


def parse_me():
    ''' Returns the current application name compontens
    (e.g. ['camagic', 'source', 'epics'])
    '''
    return os.path.basename(sys.argv[0]).split('-')

    
def parse_apps():
    ''' Separates command line arguments in sub-applications.

    The idea is that the top-level camagick calls sub-applications which
    have a name that begins with 'camagick-'. There are 3 types of
    subapplications:
    
      - Data sources (begin with 'camagick-source-...'), which create
        datasets i.e. fill the stash, but don't consume

      - Data sinks (begin with 'camagick-display-...'), which don't
        create any datasets, just view/consume the stash non-destructively
        (i.e. don't remove data)

      - Filters; these may begin with any word ('camagick-...'), and they
        take the stash and modify it (i.e. add new datasets).

    To stay modular we don't know anything about the applications and their
    command lines, but we somehow still wish to separate command line parameters
    from one another. To do this, we move from subapplication word marker
    to word marker (i.e. all display and source applications we can find
    in $PATH, and then all filters we can find in $PATH).

    This function returns an ordered dictionary as
    `{'name': ('application', (parameters...)), ...}`
    '''
    
    me = parse_me()

    # There's a general pipe structure here, but the "source" and "display"
    # are special in the sense that they're starting points / end points of
    # the pipe and require specific kinds of synchronisation.
    filters  = find_applications(me[0], exclude=['source', 'display'])

    specials = {
        'source': find_applications('%s-source' % me[0]),
        'display': find_applications('%s-display' % me[0])
    
    
    if len(sys.argv) < 2:
        print("Usage:\n\n"
              "    %s source [source... [source...]]\n"\
              "            [filter [filter... [filter...]]]\n"\

        if len(specials['source']):
            print("Available sources ('%s source <name>' for help):\n\n" % me[0],
                  *['   %s\n' % s for s in specials['source']])

        if len(specials['display']):
            print("Available displays ('%s display <name>' for help):\n\n" % me[0],
                  *['   %s\n' % d for d in specials['display'])
            

        if len(filters):
            print("Available filters ('%s <name>' for help):\n\n" % me[0],
                  *['   %s\n'%c for c in apps])

    sub_command = None
    sub_params = []
    applications = OrderedDict()
    
    for arg in sys.argv[1:] + ['']:

        # Try to find argument in any of the sub-applications lists
        if arg in [k for k in specials] + [f for f in filters] + ['']:

            # Store the previous sub-command invocation
            if sub_command is not None:
                try:
                    # reconstruct <special>-<command> invocation
                    tmp = specials[sub_command]

                    try:
                        app = tmp[sub_params[0]]
                    except KeyError:
                        logging.error("Invalid %s: %s" % (sub_command, sub_params[0]))
                        return None
                    
                    sub_command += "-"+sub_params[0]
                    sub_params = sub_params[1:]
                    
                except KeyError as e:
                    # regular <filter> invocation
                    app = filters[sub_command]

                applications[sub_command] = (app, sub_params)
                applications.move_to_end(sub_command)

            sub_command = arg
            sub_params = []
        else:
            sub_params.append(arg)

    return applications




__all__ = [
    "make_probe"
]

class dict_from_str(dict):
    # Dirty way to convert a command-line string (foo=var,foo2=var2,...)
    # into a dict.
    def __init__(self, from_str):
        super().__init__()
        s = from_str

        if s[0] == '"' and s[-1] == '"':
            raise ValueError(f'refusing to parse quoted string')
       
        while True:
            i = s.find(',')
            if (i < 0):
                if len(s) == 0:
                    break
                else:
                    i = len(s)

            param = s[:i]

            j = param.find('=')
            if j < 0:
                raise ValueError(f'not a dict: {from_str}')
            key = opcast(param[:j])
            val = opcast(param[j+1:])

            self[key] = val

            s = s[i+1:]


class strip_str(str):
    def __new__(cls, s):
        if s[0] == '"' and s[-1] == '"':
            return super().__new__(cls, s[1:-1])
        else:
            return super().__new__(cls, s)
 

def opcast(s, type_sequence=None):
    '''
    Opportunistic cast.

    Tries to convert `s` into a "higher quality"
    data type by attempting a cast to the types in `type_sequence`.
    If a `ValueError` is received, subsequent types are tried
    until the list is empty.
    If none matches, return `None`.
    The final element should be a `str` if you want to avoid `None`
    casts.    

    Args:
        s: the data, preferrably as string
        type_sequence: tuple of types to try, by default `(int, float, str)`.

    Returns: the same data as `s` just with a "better" type (if successful),
      or `None` if no conversion could be done.
    '''

    if type_sequence is None:
        type_sequence = (int, float, str)

    for Tp in type_sequence:
        try: return Tp(s)
        except (ValueError, TypeError): continue

    return None


class ValueDisconnected(Exception): pass


class _value_probe:
    def __init__(self, tag):
        self._tag = tag

class _EpicsValue_ophyd(_value_probe):
    rank = 0
    pref = '€'
    def __init__(self, pv):
        self._pv = pv[1:] if pv[0] == '€' else pv

        # pyepics apparently stumbles over empty strings. One possibility
        # how we deal with it is to query the string directly, and try to
        # do our own poor-man's conversion.
        from ophyd.signal import EpicsSignalRO
        self._sig = EpicsSignalRO(self._pv, name="value", string=True)
        self._latest_value = None
        
    def __call__(self):
        self._latest_value = opcast(self._sig.get())
        return self._latest_value


    def __repr__(self):
        s = f'€{self._pv}'
        if self._latest_value is not None:
            s += '{'+f'{self._latest_value}'+'}'
        return s


class _EpicsValue_caproto(_value_probe):
    rank = 0
    pref = '€'
    def __init__(self, pv):
        # This only works within an asyncio loop.
        self._pv = pv[1:] if pv[0] == '€' else pv

        from ophyd.signal import EpicsSignalRO
        self._sig = EpicsSignalRO(self._pv, name="value", string=True)
        self._latest_value = None
        
    def __call__(self):
        tmp = self._sig.get()
        try:
            self._latest_value = int(tmp)
        except ValueError:
            try:
                self._latest_value = float(tmp)
            except ValueError:
                self._latest_value = str(tmp)
        return self._latest_value


    def __repr__(self):
        s = f'€{self._pv}'
        if self._latest_value is not None:
            s += '{'+f'{self._latest_value}'+'}'
        return s


EpicsValue = _EpicsValue_ophyd


class EnvValue(_value_probe):
    rank = 0
    pref = '$'
    def __init__(self, env):
        self._env = env

    def __call__(self):
        from os import environ


class Value(_value_probe):
    rank = 0
    pref = None
    def __init__(self, val):
        if val in ["true", "false"]:
            self._v = val == "true"
        else:
            try:
                self._v = int(val)
            except ValueError:
                try:
                    self._v = float(val)
                except ValueError:
                    self._v = str(val)


    def __call__(self): return self._v
    def __repr__(self): return str(self._v)


class _unary_probe:
    def __init__(self, other):
        self._o = other
    def __repr__(self):
        return f"{self.marks[0]}({self._o})"

class Not(_unary_probe):
    rank = 6
    marks = [ "!", "-n" ]
    def __call__(self): return not self._o()

class Inv(_unary_probe):
    rank = 10
    marks = [ "-", "-i" ]    
    def __call__(self): return  -1*self._o()


class Empty(_unary_probe):
    rank = 8
    marks = [ "^", "-z" ]
    def __call__(self):
        print(f'Empty probe: {self._o()}')
        return len(self._o()) == 0

    
class _binary_probe:
    def __init__(self, l, r):
        (self._l, self._r) = (l, r)

    def __repr__(self):
        return f"({self._l}{self.marks[0]}{self._r})"


class Eq(_binary_probe):
    rank = 8
    marks = [ "==", "-eq" ]
    def __call__(self): return self._l() == self._r()


class Neq(_binary_probe):
    rank = 8
    marks = [ "!=", "-ne" ]
    def __call__(self): return self._l() != self._r()

class Lt(_binary_probe):
    rank = 8
    marks = [ "<", "-lt" ]
    def __call__(self): return self._l() < self._r()

class Leq(_binary_probe):
    rank = 8
    marks = [ "<=", "-le" ]
    def __call__(self): return self._l() < self._    
    
class Gt(_binary_probe):
    rank = 8
    marks = [ ">", "-gt" ]
    def __call__(self): return self._l() < self._r()

class Geq(_binary_probe):
    rank = 8
    marks = [ ">=", "-ge" ]
    def __call__(self): return self._l() < self._r()

class And(_binary_probe):
    rank = 4
    marks = [ "&&", "-a" ]
    def __call__(self): return (self._l() and self._r())


class Or(_binary_probe):
    rank = 2
    marks = [ "||", "-o" ]
    def __call__(self): return (self._l() or self._r())


import inspect

#
# building all_probes dictionary (look up all _binary_probe and
# _unary_probe classes in this module).
# 
all_probes = {}
_probes = {
    k:v for k, v in globals().items() \
    if (inspect.isclass(v) and \
        (issubclass(v, _unary_probe) or \
         issubclass(v, _binary_probe)) and \
        v not in (_unary_probe, _binary_probe))    
}
for k,v in _probes.items():
    for m in v.marks:
        all_probes[m] = v

#
# Similar to all_probes, building this dynamically.
# Typically, values are prefixed with a magic character
# (€ for EPCIS PVs, $ for shell env vars, ...). Data which
# doesn't have any prefix is used verbatim, as constants).
#
all_values = {
    v.pref:v for k,v in globals().items() \
    if (inspect.isclass(v) and \
        issubclass(v, _value_probe) and \
        (v not in (_value_probe,)))
}


def make_probe(args):
    '''
    Builds a probe from an arguments iterator.
    '''
    return consume_probe(args.copy())


def consume_probe(args, context=None, greed_rank=None):
    '''
    Recursively builds a probe expression from an arguments list.
    Modifies the list to jump past the probe.

    This is a low-level function, don't use it. Use `make_probe()`
    instead.

    Args:
        args: list of strings, one Value or Operator each.
        context: if specified, it's an l-value (context) for binary
          operators.
        greed_rank: if specified, it's the rank of the operator
          that produced `context`. Used to determine operator
          precedence.

    Returns: the top-level probe (i.e. the one that evaluates to
    all the expressions in `args`.
    '''

    # DON'T TOUCH THIS.

    # NO, really. DON'T.

    # If you think you have to, here's some useful info.
    # Operator priority generally works like this:
    #
    # Operator ranking (i.e. higher precedence):
    #  - 10: unary modifyer (- a.k.a. INV)
    #  -  8: binary comparison (==, !=, <, >, <=, >=)
    #  -  6: unary boolean (! a.k.a. NOT)
    #  -  4: binary boolean && (a.k.a. AND)
    #  -  2: binary boolean || (a.k.a. OR)
    #
    # The idea now is that we walk the "args" list and try to evaluate
    # each operator we encounter. At any point for any operator there's
    # always the _next_ item that it operates on (rval), and the _previous_
    # one (i.e. lval, we call it "context", possibly `None`) item that can
    # be used.
    #
    # We act like this is true for unary and binary operators likewise,
    # it's only that unary operators simply don't care for
    # the previous item (i.e. tose act only on the rval, regardless
    # of the value or existence of an lval).
    #
    # Precedence is a question of where we delimit the "next item":
    #
    #  - if the current operator is lower ranked than the one that
    #    comes after it, we need to be "greedy", i.e. proceed to
    #    evaluating that first before we return a value.
    #    Example: ... || 2 == ...    
    #
    #  - if the current operator is higher-or-equal-ranked, we need
    #    to return the value without sub-processing following operators.
    #    Example: ... == 2 && ...    
    #
    # As such, we have a `greed_rank` parameter that we pass on
    # recursively, as we look for the value (rval) to operate on.

    if len(args) == 0:
        # This is typically the exit point; eventually we always land
        # here by the end of the evaluation recursion. The actual result
        # is then our lvalue context.
        return context

    class NotAnOperator(Exception): pass
    
    a = args[0]
    
    try:
        Op = all_probes.get(a, None)
        if Op is None:
            raise NotAnOperator()

        # This is an operator.
        #
        # Any way we look at it, the element in args[] is the one this
        # operator applies to; i.e. this operator's rval.
        #
        # If we already have a `context`, that's our lval.
        #
        # `greed_rank` indicates how we have to compute rval:
        #
        #   - If previous operation's rank (i.e. the one that produced the
        #     `context` value) is higher than our own, then we're
        #     breaking the evaluation link
        #
        # If current operation has lower rank than previous one
        # (which generated the data in `context`), then we're at
        # the end of recursion and we need  Op(context, value)
        # as one final result.
        
        prev_rank = greed_rank if greed_rank is not None else 0

        if prev_rank < Op.rank:
            # This is the case if prev-rank is lower: this operator
            # evaluates to Op([context,] rval)
            del args[0]
            lval = context
            rval = consume_probe(args, context=None, greed_rank=Op.rank)
            if issubclass(Op, _unary_probe):
                this = Op(rval)
            elif issubclass(Op, _binary_probe):
                this = Op(l=context, r=rval)
            else:
                raise RuntimeError(f'near={a} msg="Expected operator"')

            # Continue processing to the rest of `args`, use the result so far
            # (stored in `this`) as lval context.
            return consume_probe(args, context=this, greed_rank=None)

        else:
            # This is the case of prev-rank is same or higher: this operator
            # evalueates to its lval.
            return context


    except NotAnOperator:
        
        # This is a value; whether we return the value as-is, or start
        # evaluating the arguments chain, depends on whether the operator
        # _after_ this value (consume_probe()) is stronger than the one
        # before (`greed_rank`).
        #
        # This is a decision we can't meet here, we need to defer to sub-calls
        # of consume_probe(), because we don't know future ranks -- only
        # the current and the previous (?)
        
        Val = all_values.get(a[0], all_values[None])
        del args[0]
        
        if greed_rank is None:
            # This is an l-value; we start anew at the lowest rank.
            # We pass our own value as context.
            return consume_probe(args, context=Val(a), greed_rank=Val.rank)
        else:
            # This is an r-value, we've inherited the OP rank in `greed_rank`.
            # (Why are we passing a context?)
            # Example ... _ ==  2  && _ ...   vs.   ... ||  2  == ...
            #                  ^^^                         ^^^
            # This is what it's about, vs               
            # This is where rank should play a role:
            #   - if prev-rank is same or higher, this should be a plain value "2"
            #     (eval stops there?)
            #   - if prev-rank is lower, the rest of _args_ need to be evaluated
            #     together with "2"
            return consume_probe(args, context=Val(a), greed_rank=greed_rank)

    raise RuntimeError(f'You\'re not supposed to be here with "{args}"')



import pytest, asyncio, pprint

from camagick.executor import *

from camagick import source, sink

from schema import SchemaError, SchemaUnexpectedTypeError, Optional

@pytest.fixture
def example_args():
    return (
        '--from-epics', 'prefix=KE:threshold_1:', 'image=image', 'guide=KE:acquire==0', 'width=asize1', 'height=asize2',
        '--reshape', 'shape=1,512,1028',
        '{',
        
          '--to-tiff', 'pathfmt=/home/specuser/data/image-{seq}.tif',
        
          '[',
        
            '{',
              '[',
                '--slice', '0', '10:490', '520:550',
                '{',
                  '--npfun', 'max',
                  '--npfun', 'sum',
                '}',
              ']',
              '--npfun', 'sum',
              '--npfun', 'max',
              '--npfun', 'min',
            '}',
        
            '{',
              '--to-ioc',
              '--to-summary',
            '}',
        
          ']',
        
        '}'
    )

@pytest.fixture
def example_spec2():

    return {
        "chain": [{"from": "epics",
                   "params": {
                       "image": "KE:threshold_1:image",
                       "guide": "KE:acquire==0",
                       "width": "KE:threshold_1:asize1",
                       "height": "KE:threshold_1:asize2",}},

                  {"via": [ "demote", "width", "height" ] },
                  
                  {"via": "reshape",
                   "params": {"shape": [1, 512, 1028]}},
                  
                  {"fanout": [{"to": "tiff",
                               "params": { "pathfmt": "/home/specuser/data/image-{seq}.tif"} },

                              {"chain": [{"fanout": [{"chain": [{"via": [ "slice", "0,10:490,520:550"  ],},
                                                                {"fanout": [{"via": "npfun", "params": { "op": "max" },},
                                                                            {"via": "npfun", "params": { "op": "sum" },},],}, ], },

                                                     {"via": [ "npfun", "sum" ], },

                                                     {"via": "npfun", "params": { "op": "max" },},

                                                     {"via": "npfun", "params": { "op": "min" },},], },

                                         {"fanout": [{"to": "ioc"},
                                                     {"to": "summary"},], },], }, ], }, ],
    }


def test_sequence():
    print(Sequence(int, float).validate([23, 2.4]))

    with pytest.raises(SchemaSequenceTooShortError):
        print(Sequence(int, float).validate([23]))

    with pytest.raises(SchemaUnexpectedTypeError):
        print(Sequence(int, float).validate([2.4]))

    with pytest.raises(SchemaUnexpectedTypeError):
        print(Sequence(int, float).validate([2.3, 24]))

    with pytest.raises(SchemaSequenceTooLongError):
        print(Sequence(int, float).validate([23, 3.14, 1]))

    print(Sequence(int, float, Optional(int)).validate([23, 3.14]))

    print(Sequence(int, float, Sequence.VarArguments([object]))\
          .validate([12, 3.4, "moo", 6, True]))


def test_mktype():
    assert make_pipe_type("source", "ioc") == source.ioc.Processor
    assert make_pipe_type("source", "camagick.source.ioc") == source.ioc.Processor
    assert make_pipe_type("source", "camagick.source.ioc:Processor") == source.ioc.Processor

    # "kind" not important when using dotted notation
    assert make_pipe_type(None, "camagick.source.ioc") == source.ioc.Processor


def test_validator():
    PipeValidator()


def test_with_pipe(example_spec2):
    
    print()
    with_pipe([example_spec2], print)

    pm = PipeMaker()

    with_pipe([example_spec2], pm)

    print('Stack:', pm)
    

def test_validate(example_spec2):

    scm = PipeValidator()

    r1 = scm.validate({
        "chain": [
            { "from": [ "epics", "moo" ] },
            { "via": "npfun", "params": { "op": "sum" } },
            { "via": "npfun", "params": [ "sum" ] },
            { "via": [ "npfun", "sum" ] },
            { "via": [ "demote", "moo" ] },
            { "to": "summary" },
        ]
    })

    assert len(r1) > 0

    r2 = scm.validate({
        "fanout": [
            { "from": "ioc", "params": { "prefix": "fum" } },
            { "chain": [
                { "via": "npfun", "params": [ "sum" ] },
                { "to": "summary" },],},
            { "to": "ioc" },
        ]        
    })

    assert len(r2) > 0

    r3 = scm.validate(example_spec2)


def test_args(example_args):
    spec = pipe_spec_from_args(example_args)

    pprint.pprint(spec)



from camagick.processor import ProcessorBase, SinkBase
import numpy as np
import zarr, logging, asyncio

logger = logging.getLogger(__name__)

class RefuseIndex(RuntimeError): pass

class RefuseOverwrite(RefuseIndex): pass

class RefuseSparse(RefuseIndex): pass


class ZarrDataStacker:
    '''
    Simple Zarr stacking-based storage.
    '''
    
    def __init__(self, store, group=None, mode=None, zarr_args=None):
        '''
        Args:
            store: Zarr store of where to put the data.
        
            group: Path _inside_ the HDF5 file where to save the data.

            mode: What to do when we're required to save a data point
              in a dataset for which there's already data available
              in the same dataset. See `Processor.__init__()` for documentation
              on available modes.
        '''        
        self._zstore = store
        self._zgroup = group if group is not None else ''
        self._zargs = zarr_args if zarr_args is not None else {}
        self._frame_mode = mode if mode is not None else 'a+'

        logger.info(f'msg="Zarr storage" store="{store}" group="{group}"')


    def _ensure_dataset(self, znode, name, dtype, point_shape):
        '''
        Received with the first data point, sets names, types etc.

        Args:
            znode: Zarr group object
            name: name of the dataset
            dtype: data type (numpy-dtype)
            point_shape: shape of the data point; note that
              the actual shape of the resulting data set will contain
              more dimension
        '''

        dshape = (0, *point_shape)

        fname = str(znode.store_path)
        
        try:
            dset = znode[name]

            if (dset.dtype != dtype):
                if 'O' not in self._frame_mode:
                    raise RuntimeError(f'msg="Dataset exists with different dtype"'
                                       f'have="{dset.dtype}" want="{dtype}" file="{fname}"')
                else:
                    raise KeyError()

            if dset.shape[1:] != dshape[1:]:
                if 'O' not in self._frame_mode:                
                    raise RuntimeError(f'msg="Dataset exists with different shape"'
                                       f'have="{dset.shape}" want="{dshape}" file="{fname}"')
                else:
                    raise KeyError()

        except KeyError:
            logger.info(f'msg="Creating dataset" file="{fname}" name="{name}" '
                        f'shape="{dshape}" dtype={dtype}')
            znode.create_array(name=name, shape=dshape, dtype=dtype, overwrite=True)


    def _ensure_frame(self, dset, index, fname):
        
        idiff = (index-dset.shape[0]) + 1
        
        if idiff == 1:
            shape_diff = 1
            
        elif idiff > 1:
            if '+' in self._frame_mode:
                shape_diff = idiff
            else:
                raise RefuseSparse(f'msg="Sparse writing not allowed; use top-up `+` flag" '
                                   f'index={index} frames={dset.shape[0]} dataset="{dset.name}" '
                                   f'file={fname}')
            
        elif idiff <= 0:
            if 'a' in self._frame_mode:
                raise RefuseOverwrite(f'msg="Overwrite not permitted" index={index} file={fname} '
                                      f'frames={dset.shape[0]} dataset="{dset.name}" '
                                      f'file=""')

            elif 'o' in self._frame_mode or 'O' in self._frame_mode:
                shape_diff = 0

            elif 'x' in self._frame_mode:
                logger.info(f'msg="Resetting dataset" index={index} file={fname} '
                            f'frames={dset.shape[0]} dataset="{dset.name}"')
                new_shape = [int(i) for i in (0, *(dset.shape[1:]))]
                dset.resize(new_shape)

                shape_diff = index+1

            else:
                raise RuntimeError(f'BUG: No shape diff for mode: {self._frame_mode}')

        if shape_diff > 0:
            new_shape = [int(i) for i in (dset.shape[0]+shape_diff, *(dset.shape[1:]))]
            logger.debug(f'msg="Resizing dataset" ' #'file="{dset.group.filename}" '
                         f'name={dset.name} size={dset.shape[0]} '
                         f'index={index} diff={shape_diff} idiff={idiff} '
                         f'shape={new_shape}')
            dset.resize(new_shape)


    def push_data(self, index, **data):
        '''
        Received for every data point in a scan.

        Args:
            index: Integer to determine the position of the current
              data point in the overall (larger) dataset. This should
              always point to the next available index in the first
              dimention, i.e. if the existing dataset has dimension
              (37, N, M), with the current data point to become
              the 38th, then `index` needs to be 37 (counting starts
              at 0).

              Behavior if the index doesn't match the shape is defined
              by the class-global parameter `mode`.

              If this is `None`, the next free slot (i.e. current_size+1)
              is used, essentially making the operation always succeed
              as if "append" mode were active.
              
            **data: dictionary from data name ("tag") to data arrays
        '''

        znode = zarr.open_group(store=self._zstore, path=self._zgroup,
                                mode='a', **self._zargs)

        #print(store, znode)
            
        for key, container in data.items():
            dtype = type(container.flatten()[0].item())
            self._ensure_dataset(znode=znode,
                                 name=key,
                                 dtype=dtype,
                                 point_shape=container.shape)

            dset = znode[key]

            if index is None:
                index = dset.shape[0]

            self._ensure_frame(dset, index, fname=self._zstore)
            item = np.array(container)
            dset[index] = item



class Processor(SinkBase):
    '''
    Saves data as a dataset in a Zarr structure.

    The sink tries to make it easy to obtain uniform, synchronous,
    well-defined arrays of data of the same length. It does so
    mostly by filling up empty frame slots when there isn't any
    data available, or by overwriting data if necessary, according
    to the user specification. This is governed according to the
    `mode` initialization parameter, which determines what to
    do when a dataset already exists.
    '''
    
    def __init__(self,
                 store: str,
                 mode: str = 'a+',
                 index: str = '',
                 strict: bool = True,
                 mt: bool = True,
                 **zarr_args
                 ):
        '''
        Args:
            path: the path format. This is expected to be of the format
              "<file>#<group>/[<dataset>]". The processor splits apart
              the outer part ("<file>") and inner part ("<group>", i.e.
              "inside" the Zarr folder. This isn't of much importance
              when opening already existing zarr groups, but does matter
              if a group doesn't exist yet and needs to be created.
              The splitting is done, as a feature of the Caspy processor,
              at the first '#' character. As such, the '#' character
              never reaches lower Zarr library levels.

              Note that the <file> part is passed directly to
              `zarr.open_group()` as the `store=...` parameter, which means
              that you can make use of Zarr's vast and diverse fs-transparent
              load mechanism by supplying full URLS (e.g. "file:///..." or
              "s3://..." or "http://...").

              If the name of the data set (i.e. "<dataset>") is missing,
              then the store needs to end on "/". In that case, the processor
              will append a dataset name based on the tag.

            mode:  Must be one of `a` (append), `o` (overwrite) or
              `x` (reset), possibly in combination with `+` (top-up):
        
                 - `a`: require that the data point be always appended
                   to the current dataset -- never overwrite or delete
                   existing data.
        
                 - `o`: overwrite already existing data points (frames).

                 - `O`: same as `o`, but additioally delete and re-create
                   whole dataset entries, if the shape or data type of the
                   existing dataset does not match the new request.

                 - `x`: reset the entire dataset (i.e. delete all existing
                   data points) if a data point already exists at the
                   specified index.

                 - `+`: top-up marker will fill up empty slots, but without
                    actually writing any data to them. They will remain
                    filled with default HDF5 settings according  to the data
                    type. If top-up is not specified, otherwise the operation
                    will fail if index is higher than the one of the next
                    slot in turn.
        
            index: name of the context variable which contains the current
              frame index (preferrably integer). If this is an empty string,
              no index is used and essentially every operation is an "append".

            strict: if this is `True` (the default), then any saving errors
              being raised will be allowed to bubble up to the pipe processor.
              Typically, this means the application is being shut down.
              When `False`, then `RefuseIndex` type of errors (typically when
              a frame index is out of place) are being reported, but ignored.

            mt: if set to `True` (the default), writing individual data sets
              to Zarr will happen in parallel, each in its own thread. Given
              that Zarr is built to handle this exact scenario, and we're
              typically writing very small data amounts but with a significant
              amount of API I/O overhead, this is almost always what you want.

            **zarr_args: any arguments to be passed on to `zarr.open_group()`
              for storage.
        '''
        super().__init__()
        
        self._store_fmt = store
        self._store_kw = zarr_args
        self._index_ctx_tag = index
        self._stacker_mode = mode
        self._current_file_and_node = (None, None)
        self._strict = strict
        self._threaded = mt


    def store_inner_split(self, full):
        i = full.find('#')
        if i > 0:
            zfile, zlocal = full[:i], full[i+1:]
        elif i < 0:
            zfile, zlocal = full, '/'
        else:
            raise RuntimeError(f'msg="Don\'t understand path#group combo" store="{full}"')

        return zfile, zlocal


    def group_dset_split(self, inner, default_dset=None):
        l = inner.rfind('/')
        if l < 0:
            raise RuntimeError(f'msg="Cannot determine dataset name from inner path" '
                               f'path="{inner}"')

        node_path = inner[:l+1]
        dset_path = inner[l+1:]

        if len(dset_path) == 0:
            dset_path = default_dset

        return node_path, dset_path


    def get_index(self, context):
        if self._index_ctx_tag not in (None, ''):
            try:
                if self._index_ctx_tag[0] == '@':
                    _index = context[self._index_ctx_tag[1:]]
                else:
                    _index = context[self._index_ctx_tag]
            except KeyError as e:
                err = \
                    f'msg="Index key not available in this frame" ' \
                    f'expected="{self._index_ctx_tag}" ' \
                    f'available="{','.join(context.keys())}"'
                raise RefuseIndex(err)
        else:
            _index = None

        return _index


    def handle_index_error(self, err, tag=None):
        if self._strict or (not isinstance(err, RefuseIndex)):
            raise err
        logger.error(err)
        logger.error(f'msg="Ignoring frame index error as instructed" tag={tag}')

        

    def push_frame(self, tag, cont, index, context, store_fmt, return_exceptions=True):
        full_path = store_fmt.format(tag=tag, **context)
        zfile, zlocal = self.store_inner_split(full_path)
        node_path, dset_path = self.group_dset_split(zlocal, tag)

        try:
            stor = ZarrDataStacker(store=zfile, group=node_path,
                                   mode=self._stacker_mode,
                                   zarr_args=self._store_kw)
            stor.push_data(index=index, **{dset_path: cont})
            return (tag, f'{full_path}[{index}]')
        
        except Exception as e:
            if return_exceptions:
                return e
            else:
                raise e


    async def __call__(self, data=None, context=None):
        
        if data is None:
            data = {}
        if context is None:
            context = {}


        try:
            _index = self.get_index(context)
        except RefuseIndex as e:
            self.handle_index_error(e)
            logger.warning(f'msg="Skipping frame altogether"')
            return


        if self._threaded:
            result = await asyncio.gather(*[
                asyncio.to_thread(
                    self.push_frame, tag, cont, _index, context, self._store_fmt
                ) for tag, cont in data.items()
            ])
        else:
            result = [
                self.push_frame(tag, cont, _index, context, self._store_fmt) \
                for tag, cont in data.items()
            ]

        ret = {}
        for r in result:
            if isinstance(r, Exception):
                self.handle_index_error(r)
            else:
                ret[r[0]] = r[1]

        return ret



#!/usr/bin/pyton3

''' Various stash access helpers '''

import fcntl, asyncio, logging, xarray, random, time

class ZarrStash:
    ''' Stashes the data in a Zarr archive on disk.

    Synchronizes seed/crop access using lock files.
    '''

    def __init__(self, storage, lockKey, data=None):
        ''' Intializes stash system.

        Args:
            storage: Storage path (zarr folder)
            lockKey: Key detail for locking. This is used to generate
              a unique lock file based on the storage path.
            data: (optional) xarray data to store.
        '''
        self.storage = storage
        self.lockKey = lockKey
        self.lockPath = storage+lockKey+".lock" if lockKey is not None else None
        self.lockFile = None
        self.myLock = False

        self.contextCnt = 0


    def tryLock(self):
        ''' Attempts to lock, returns True on success or False otherwise. '''

        # disable lock of key is None
        if self.lockKey is None:
            return True
        
        try:
            self.lockFile = open(self.lockPath, 'w+')
            fcntl.flock(self.lockFile.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
            self.myLock = True
            return True
        except FileNotFoundError:
            raise
        except IOError as e:
            return False
        

    def unlock(self):
        ''' Frees the lock. '''
        if self.lockKey:
            return
        
        fcntl.flock(self.lockFile.fileno(), fcntl.LOCK_UN)
        self.lockFile = None
        self.myLock = False

    def nolockSave(self, data, keys=None):
        data.to_zarr(store=self.storage, mode='a')

    def nolockLoad(self, keys=None):
        return xarray.load_dataset(self.storage, engine='zarr')        
        
    def save(self, data, keys=None):
        with self:
            return self.nolockSave(data, keys)

    def load(self, keys=None):
        with self:
            return self.nolockLoad(keys)

    def __enter__(self):
        self.contextCnt += 1
        self.waitLock()
        return self

    def __exit__(self, exc_type, exc, tb):
        if self.contextCnt > 1:
            self.contextCnt -= 1
        if self.contextCnt == 0:
            self.unlock()

    async def waitLock(self, pollPeriod=1e-5):
        ''' Asynchronously wait for a lock on the stash.
        '''
        while not self.tryLock():
            time.sleep(pollPeriod+pollPeriod*random.random()*0.01)
    

    async def __aenter__(self):
        self.contextCnt += 1
        await self.asyncWaitLock()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self.contextCnt > 1:
            self.contextCnt -= 1
        if self.contextCnt == 0:
            self.unlock()    
    
    async def asyncWaitLock(self, pollPeriod=1e-5):
        ''' Asynchronously wait for a lock on the stash.
        '''
        while not self.tryLock():
            await asyncio.sleep(pollPeriod+pollPeriod*random.random()*0.01)
    
    async def asyncSave(self, data, keys=None):
        async with self:
            self.nolockSave(data, keys)
            
    async def asyncLoad(self, keys=None):
        async with self:
            return self.nolockLoad(keys)



import asyncio, contextlib, pprint, parse, logging

from camagick.processor import ProcessorBase, SourceBase
from camagick.probe import opcast
from camagick.ioc import IocServer, InstaInputPVDB

from caproto.server import server as ca_server

import numpy as np

logger = logging.getLogger(__name__)

class PvMapper:
    pass

class Processor(SourceBase):

    def __init__(self,
                 prefix: str = '',
                 hang: bool = True,
                 suppress: str = None,
                 **pvspec):
        '''
        Initializes an IOC-source processor.

        This one spans a number of PVs for external users / components to write
        data into.

        Args:
        
            prefix: the IOC prefix, will be prepended (verbatim, no trailing ":")
              to every PV.

            hang: if set to any of `True`, `1`, `"yes"`, `"any"` the processor will
              not pass on data until new data has explicitly been written to any of
              the variables. If set to `"no"`, `0`, `False`, the Processor will pass
              on data every time it is asked.

            suppress: One of "initial", "same", or `None` (the default). If set
              to "initial", the initial PV value -- one that wasn't explicitly set
              externally after startup -- won't be propagated, and the IOC will
              instead wait forever until at least one value is explicitly written.
              "same" will suppress propagation values that are the same as the value
              the PV already holds.
        
            **pvspec: mapping the data label (keys) to PVs (sans prefix) and optionally
              starting values, in the format `'label' -> 'PV:string(val)'`, or
              `{ 'label': 'PV:string' }`.
        '''

        # Extract the PV labels, names and initial values. If there's no initial value,
        # we use a float (0.0). This is a { label: (pvname, value) } map.
        self._pv_value_map = self._parse_pv_initial(pvspec)
        self._suppress = suppress.split(',') if suppress is not None else []

        # This is how we map things locally: full-pv <-> short name
        self._pv_map = {
            f'{prefix}{v[0]}':k \
            for k,v in zip(pvspec.keys(),self._pv_value_map.values())
        }
        self._pv_unmap = {}
        for k,v in self._pv_map.items():
            if v in self._pv_unmap:
                raise RuntimeError(f'Double mapping: {k} <-> {v}')
            self._pv_unmap[v] = k

        #logger.debug(f'map: {self._pv_map}')
        #logger.debug(f'unmap: {self._pv_unmap}')

        pvmap = { f'{prefix}{v[0]}':v[1] for k,v in self._pv_value_map.items() }
        self.ioc = InstaInputPVDB(pv_map=pvmap,
                                  hang=self._check_true(hang),
                                  suppress_initial=('initial' in self._suppress),
                                  suppress_same=('same' in self._suppress),
                                  verify=self._verify_pv_value)
        self.ioc_server = IocServer(self.ioc)


    async def _verify_pv_value(self, pvname, value):
        return await self.verify(self._pv_map[pvname], value)


    async def verify(self, tag, value):
        '''
        Called for every incoming value.

        If this raises an exception, the value is not propagated
        through the CA internals. (Typically `caproto.CaprotoValueError`?)
        If a modified value is returned, the modified value is
        propagated instead.
        '''        
        pass


    def _check_true(self, hang):
        if hang in (1, True):
            return True
        elif isinstance(hang, str):
            return hang.lower() in ("yes", "true", "any")


    def _parse_pv_initial(self, label_spec_dict):
        ret = {}
        for label, spec_param in label_spec_dict.items():
            spec = str(spec_param)
            result = parse.parse('{name}({value})', spec)
            if result is not None:
                ret[label] = (result['name'], opcast(result['value']))
            else:
                ret[label] = (spec, np.nan)
        return ret

    
    async def startup(self):
        await self.ioc_server.startup()


    async def shutdown(self):
        await self.ioc_server.shutdown()

    
    async def __call__(self, data=None, context=None):
        await asyncio.gather(self.ioc.wait_for_input(), return_exceptions=False)
        my_data = {}
        for k,v in self.ioc._pvdb.items():
            if isinstance(v, ca_server.PvpropertyString):
                my_data[self._pv_map[k]] = v.value[0]
            else:
                v = v.value \
                    if isinstance(v.value, str) or hasattr(v.value, "shape") \
                       else np.array(v.value)
                my_data[self._pv_map[k]] = v

        #if self._transparent and data is not None:
        #    my_data.update(data)

        return my_data 



async def test_run():
    p = Processor(prefix='Jinkies:', hang=False,
                  scan='Scan(3.14)', stan='Stan(34)')
    await p.startup()

    try:
        while True:
            data = await p()
            print(data)
            
    finally:
        await p.shutdown()
    

if __name__ == "__main__":
    asyncio.run(test_run())



from camagick.storage.base import ArrayAccess as ArrayAccessBase
from camagick.storage.base import RefuseIndex, RefuseOverwrite, RefuseSparse

import numpy as np
import h5py, logging, os

logger = logging.getLogger(__name__)    

class ArrayAccess(ArrayAccessBase):
    '''
    Implementation of a HDF5 file array access.
    '''

    def __init__(self, store=None, sdir=None, mode=None, **tags):
        super().__init__(store, sdir, mode, **tags)

        
    async def _open_writable(self, store):
        self.__ensure_directory_of(store)
        return h5py.File(store, 'a')


    async def _open_readable(self, store):
        return h5py.File(store, 'r')


    def __ensure_directory_of(self, path):
        # Makes sure all path components up to the HDF5 file name exist
        print('dir of:', path)
        d = os.path.dirname(path)
        if (len(d) > 0):
            if not os.path.exists(d):
                os.makedirs(d)
                return
            if (not os.path.isdir(d)):
                raise RuntimeError(f'msg="Path is required to be a folder" '
                                   f'path="{d}"')


    def _ensure_dataset_base(self, stor, dset, **attrs):
        # Makes sure `nodepath` exists in the current h5 file and is a data group
        nodepath, dsetonly = self._split_dset_path(dset)
        subpath = ""
        subnode = stor
        for current in nodepath.split('/'):

            if len(current) == 0:
                continue # nothing to do
            try:
                subnode = subnode[current]
                
            except KeyError:
                logger.info(f'msg="Creating node" file="{subnode.filename}" '
                            f'parent="{subnode.name}" '
                            f'name="{current}"')
                subnode = subnode.create_group(current)

        return (subnode, dsetonly)


    def _dtype_spec_to_dtype(self, ds):
        if isinstance(ds, str):
            try:            
                return {
                    'number': float,
                    'integer': int,
                    'array': float,
                }[ds]
            except KeyError:
                errmsg = f'tag="{key}" msg="Don\'t know how to save" dtype="{dt_spec}"'
                logger.error(errmsg)
                raise RuntimeError(errmsg)

        else:
            return ds


    def _ensure_dataset(self, path, base, dtype=None, frame_shape=None, **attrs):
        '''
        Received with the first data point, sets names, types etc.

        Args:
            name: name or path of the dataset        
            base: HDF5 group object to contain the dataset
            dtype: data type (numpy-dtype)
            point_shape: shape of the data point; note that
              the actual shape of the resulting data set will contain
              more dimension
        '''

        if self.readonly:
            return base[path] if len(path) > 0 else base
        
        dshape = (0, *frame_shape)
        
        try:
            dset = base[path]

            if (dset.dtype != dtype):
                if  'O' not in self._frame_mode:
                    raise RuntimeError(f'msg="Dataset exists with different dtype"'
                                       f'have="{dset.dtype}" want="{dtype}"')
                else:
                    del base[path]
                    raise KeyError()

            if dset.shape[1:] != dshape[1:]:
                if  'O' not in self._frame_mode:                
                    raise RuntimeError(f'msg="Dataset exists with different shape"'
                                       f'have="{dset.shape}" want="{dshape}"')
                else:
                    del base[path]
                    raise KeyError()

        except KeyError:
            logger.info(f'msg="Creating dataset" file="{base.file.filename}" name="{path}" '
                        f'shape="{dshape}" dtype={dtype}')
            dset = base.create_dataset(path, dshape, maxshape=(None, *dshape[1:]),
                                       dtype=dtype, compression='lzf')

        return dset




from camagick.storage.base import ArrayAccess as ArrayAccessBase
from camagick.storage.base import RefuseIndex, RefuseOverwrite, RefuseSparse

import numpy as np
import h5py, logging, os

logger = logging.getLogger(__name__)

class _MemoryArrayRepo:
    def __new__(cls, *a, **kw):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls, *a, **kw)
        return cls._instance

    def __init__(self):
        self._storage = {}


class ArrayAccess(ArrayAccessBase):
    '''
    Implementation of an array store that holds all the data in memory.

    This is mainly for debugging and testing purposes, but might be
    useful. Stores the data in a number of stacked dict() containers
    -- one for each level (starting wiht '/'), and the datasets as numpy
    arrays.
    '''

    class _dset_wrapper:
        # This is needed because Numpy arrays are difficult to
        # resize in-place. We hold one level of indirection so
        # that no other instance references the actual array.
        def __init__(self, ar):
            self._ar = ar

        @property
        def shape(self):
            return self._ar.shape

        @property
        def dtype(self):
            return self._ar.dtype

        def __setitem__(self, *a):
            return self._ar.__setitem__(*a)

        def resize(self, new_shape):
            self._ar.resize(new_shape)

        def __repr__(self):
            return self._ar.__repr__()
        

    def __init__(self, store=None, sdir=None, mode=None, **tags):
        self._repo = _MemoryArrayRepo()
        super().__init__(store, sdir, mode, **tags)

    def _resize_dataset(self, dset, new_shape):
        dset.resize(new_shape)

    async def _open_writable(self, store):
        return self._repo._storage.setdefault(store, {})

    def _close(self, tag):
        pass

    def _ensure_dataset_base(self, stor, dset, **attrs):
        d = stor
        parts = dset.split('/')
        for c in parts[:-1]:
            d = d.setdefault(c, {})
        return d, parts[-1]

    def _ensure_dataset(self, path, base, dtype=None, frame_shape=None, **attrs):
        parts = path.split('/')
        name = parts[-1]
        if len(name) == 0:
            raise RuntimeError(f'bad dataset: {name}')
        if len(parts) > 1:
            for c in parts[:-1]:
                base = base[c]
                
        if parts[-1] in base:
            d = base[parts[-1]]
            if (not self.readonly) \
               and ((d.dtype != dtype) or \
                    (d.shape[1:] != frame_shape)):
                raise RuntimeError(f'incompatible array {d.dtype}/{d.shape[1:]}')
            return d
        elif not self.readonly:
            d = base[parts[-1]] = self._dset_wrapper(
                np.ndarray((0, *frame_shape), dtype=dtype)
            )
            return d
        else:
            raise RuntimeError('read-only array')
        



import asyncio, logging, importlib

logger = logging.getLogger(__name__)

def easy_load(*any_kind, sources=None, sinks=None, pipes=None, flows=None):
    '''
    Helper to conveniently load camagick processors.

    Specify sources, sinks and flow elements by their module
    name. They'll end up as items in a namespace class,
    named <module>Source/Sink/Flow.
    '''
    cat = {
        'source': sources if sources is not None else [],
        'sink': sinks if sinks is not None else [],
        'pipe': pipes if pipes is not None else [],
        'flow': flows if flows is not None else ['chain', 'fanout']
    }

    class _tmp:
        pass

    ret = _tmp()

    for c in cat:
        for mod in cat[c]:
            imported_name = mod[0].upper()+mod[1:]+c[0].upper()+c[1:]
            base = importlib.import_module(f'camagick.{c}.{mod}')
            setattr(ret, imported_name, getattr(base, 'Processor'))

    return ret


class DuplicateDataTag(RuntimeError): pass

#
# Base for all CAMagick processing components (sources, modifyers, sinks)
#

class ProcessorBase(object):

    async def startup(self):
        pass

    async def shutdown(self):
        pass
    
    async def __call__(self, data=None, context=None):
        '''
        This is the actual workhorse. It's expected to accept
        a data dictionary (if anything available), and return
        a modified dictionary.

        There's an exception for sources (which usually ignore incoming
        data), and sinks (which usually don't return anything anymore).

        Args:
        
            data: the data the processor is expected to primarily apply to.
              This is usually whatever is requested in the 'from-epics'
              class of sources.

            context: A dictionary of additionaly data (possibly also
              of 'from-epics' origin), which isn't to be primarily
              processed, but which is available for further decision
              making. Each processor is free to make use of data here
              as it sees fit.

        Returns: may return None (if it's a sink), or a set of modified
          data. The modified data may or may not contain all, any, or none
          of the original data keys, and may also create its own new data
          keys (and values).
        '''
        raise RuntimeError(f'Not implemented')


class SourceBase(ProcessorBase):
    '''
    Base class from which all source-like pipes are derived.

    This is only used within specific flow elements (Chains)
    to auto-merge results with the results of the previous pipe,
    instead of replacing them outright.
    '''
    pass


class SinkBase(ProcessorBase):
    '''
    Base class form which all sink-like pipes are derived.

    This is used within specific flow elements (Fanout) to block
    output data from propagating.
    '''
    pass


class FlowBase(ProcessorBase):
    '''
    Base class from which all flow elements are derived.

    Common functionality is .startup() / .shutdown() loops, and __repr__().
    '''
    
    def __init__(self, *pipes):
        self.subpipes = pipes


    async def startup(self):
        aw = [
            (p.startup() if hasattr(p, "startup") else _nop()) \
            for p in self.subpipes
        ]
        
        await asyncio.gather(*aw, return_exceptions=False)


    async def shutdown(self):
        async def _nop(): pass
        aw = [
            (p.shutdown() if hasattr(p, "shutdown") else _nop()) \
            for p in self.subpipes
        ]
        
        await asyncio.gather(*aw, return_exceptions=False)


    def __str__(self):
        parts = '\n'.join([str(p) for p in self.subpipes])
        return f'({self.__class__.__name__}:\n  {parts})'        
        

    def __repr__(self):
        return str(self)
        



from camagick.processor import ProcessorBase
import numpy as np

def _lucky_parser(s):
    
    if s == 'None':
        return None

    if s in ('False', 'false'):
        return False

    if s in ('True', 'true'):
        return True

    for t in [ int, float ]:
        try:
            return t(s)
        except ValueError:
            continue

    return s


class Processor(ProcessorBase):
    '''
    Numpy-based single-array operations (mean, average, max, ...)

    Essetially, this processor will call the named operation on
    every input dataset, and replace the data with the result of
    the operation.

    Hint: you may want to use the `--rename` processor right after
    this one, and in conjunction with a suitable combination of
    fanout / chain flow control, if you intend to keep both data sets.

    Example:

      caspy ... { [ --only .* ]
                  [ --npstat1 sum --rename {}_sum ] } ...

    This will calculate the sum of incoming data, and pass on the
    sum as {}_sum, as well as the original data. (The latter part is
    owing to the fact that we've enclosed the operations in a fanout
    structure.)
    '''

    def __init__(self, op: str, **args):
        '''
        Args:
            op: operation to use (as string, as available in the numpy namespace).
              Any operation is allowed that uses (a) a single array as its first
              input parameter, and (b) only parameters that can be named as its
              further parameters.
        
            **args: named arguments to pass on as the operation's parameters after
              the array.
        '''
        self.stat_proc = getattr(np, op)

        # This will fail because most 'args' values will be strings,
        # and we need to translate them to useful python types
        # (True, None, int(...), ...)
        self.stat_kwargs = { k:_lucky_parser(v) for k,v in args.items() }
    
    async def __call__(self, data=None, context=None):        
        return {
            k:self.stat_proc(v, **self.stat_kwargs) \
            for k,v in data.items()
        }
    



#!/usr/bin/python3

import logging, time, math, random, asyncio

from camagick.processor import ProcessorBase, SourceBase
from camagick.probe import opcast

import numpy as np

import tango
from tango import asyncio as tas

logger = logging.getLogger(__name__)

class Processor(SourceBase):
    '''
    Reads data as a tango client, from tango variables
    '''

    def __init__(self, dev: str, **tags):
        '''
        Produces data.

        Args:
            **tags: tag -> tango variable
        '''
        self._dev = dev
        self._tags = tags


    async def startup(self):
        self._proxy = await tas.DeviceProxy(self._dev)
        logger.info(f'msg="Have device proxy" proxy={self._proxy}')


    async def shutdown(self):
        pass


    async def __call__(self, data=None, context=None):

        if data is None:
            data = {}

        awaitables = [ self._proxy.read_attribute(self._tags[k]) for k in self._tags ]
        result = await asyncio.gather(*awaitables)

        my_data = {
            k:np.array(r.value) for k,r in zip(self._tags, result)
        }

        #my_data.update(data)

        return my_data



import logging, importlib, inspect, pkgutil, typing, types, traceback, asyncio

logger = logging.getLogger(__name__)

from functools import partial
from pprint import pprint, pformat
import numpy as np
from schema import Schema, Or, And, Use, Const, Optional, SchemaError

from camagick.flow.chain import Processor as ChainFlow
from camagick.flow.fanout import Processor as FanoutFlow
from camagick.flow.chain import Processor as ChainFlow
from camagick.probe import opcast, dict_from_str, strip_str

__all__ = [
    "PipeExecutor",
    "PipeValidator",
    "Sequence",
    "SchemaSequenceTooLongError",
    "SchemaSequenceTooShortError",
    "PipeMaker",
    "PipeHelpRequested",
    "find_processors",
    "with_pipe",
    "make_pipe_type",
    "pipe_spec_from_args",
    "QuitApplication",
]

_kind2ymlkey = {
    'sink': 'to',
    'source': 'from',
    'pipe': 'via'
}

_ymlkey2kind = { v:k for k,v in _kind2ymlkey.items() }

class QuitApplication(Exception): pass

class PipeExecutor:
    def __init__(self, flow, warn_residual: bool = False):
        self.flow = flow
        self.warn_residual = warn_residual
        self._ctx_counter = 0


    async def __aenter__(self):
        if self._ctx_counter == 0:
            await self.flow.startup()
            self._task = asyncio.create_task(self.run(startup=False))
        self._ctx_counter += 1
        return self


    async def __aexit__(self, *args):
        self._ctx_counter -= 1
        if self._ctx_counter == 0:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass


    async def step(self):
        context = {}
        out_data = await self.flow(data=None, context=context)
        if (out_data is not None) and len(out_data)>0 and self.warn_residual:
            logger.warning(f'msg="Residual data keys" keys="{out_data.keys()}"')
        return out_data

        
    async def run(self, startup=True):
        
        if startup:
            # If we came here through __aenter__, the flow object should already
            # have been initialized. (Apparently this is necessary to avoid the
            # case where we've entered the 'with' context, but the task wasn't
            # initialized yet.
            # All other cases should not use the 'startup' parameter.
            await self.flow.startup()

        try:
            while True:
                out_data = await self.step()

        except QuitApplication as e:
            logger.info(f'msg="Closing now."')

        except Exception as e:
            traceback.print_exc()
            logger.error(e)
            logger.error('Good bye')

        except KeyboardInterrupt:
            logger.info(f'msg="Good bye."')

        finally:
            logger.info(f'msg="Exector shutting down"')
            await self.flow.shutdown()


def find_processors(base=None, cls=None, full=None, raise_missing=False, noisy_missing=False):
    # Searches a module `base` for Processor submodules
    # (i.e. submodules which have a class named `cls`,
    # typically "Processor").
    #
    # As a convenience, if either `base` or `cls` are not
    # specified, they are read out from `full` first
    # (expected format: "{modbase}.{name}:{cls}")

    if base is None and full is not None:
        ## Split 'full' into its constiuent parts.
        ## Generally, there will only be 1 processor.
        parts = full.split(':')
        if len(parts)>1:
            cls = parts[-1]
        modparts = parts[0].split('.')
        if len(modparts)>1:
            base = modparts[:-2]
            candidates = [modparts[-1]]
        else:
            candidates = [modparts[0]]
    else:
        modbase_obj = importlib.import_module(base)
        candidates = [p.name for p in pkgutil.iter_modules(modbase_obj.__path__)]

    result = {}
    for pname in candidates:

        try:
            if len(base)>0:
                mod_name = f'{base}.{pname}'
            else:
                mod_name = f'{pname}'

            modobj = importlib.import_module(f'{mod_name}')

        except ModuleNotFoundError as e:
            if raise_missing:
                raise
            continue
        except Exception as e:
            if raise_missing:
                raise
            if noisy_missing:
                logger.error(f'msg="Cannot import processor, likely a bug in underlying libraries" name={base}.{pname}')
                logger.error(e)
            continue

        try:
            Cls = getattr(modobj, cls)
            #if not issubclass(Cls, Processor):
            #    continue
        except AttributeError:
            continue

        result[pname] = Cls

    return result
            


def make_pipe_type(kind, name, show_error=True):
    '''
    Returns a Python type for the specified processor.

    Args:
    
        kind: processor kind (one of "sink", "source" or "pipe").
          Currently this is only used for building the internal
          module name. If the module is external, the kind
          parameter is ignored.
    
        name: either a single module name from within `camagick.{kind}.{sub}`,
          or a full Python module name of the format `{module}[:{class}]`,
          where `{module}` may contain several components. The `:{class}`
          part can be omitted if `{module}` contains at least one dot "."
          in which case the class name `Processor` is assumed, similar
          to the built-in processors.
    '''

    mod_cls = name.split(':')

    is_internal = not len(mod_cls)>1
    
    if len(mod_cls) > 1:
        clsname = mod_cls[-1]
    else:
        clsname = "Processor"
        
    try:
        if not is_internal:
            modname = mod_cls[0]
            mod_obj = importlib.import_module(modname)
        else:
            if '.' not in name:
                modname = f'camagick.{kind}.{name}'
            else:
                modname = name
            mod_obj = importlib.import_module(modname)
        cls_obj = getattr(mod_obj, clsname)
        return cls_obj

    except Exception as e:
        if show_error:
            logger.error(f'msg="Error loading processor" name="{name}" '
                         f'module="{modname}.{clsname}')
        raise


def pipe_type_from_node_dict(node):
    key, data = next(iter(node.items()))
    kind = _ymlkey2kind[key]
    return make_pipe_type(kind=kind, name=data[0])


def _type_spec_from_arg(s):
    # Once the leading '--' is removed, 
    # Typical format is either "from-{}", "to-{}" or just plain "{}"
    # for built-in types, and "from=mod.ule[:Object]" or just
    # plain "mod.ule[:Object]" otherwise.
    parts = s.split('-')
    if len(parts) == 2:
        if parts[0] not in ('from', 'to'):
            raise RuntimeError(f'{s}: unknown node kind {parts[0]}')
        return parts
    else:
        return ("via", parts[0])


class PipeHelpRequested(Exception):
    def __init__(self, node):
        self.node = node
        k, n = next(iter(node.items()))
        self.kind = _ymlkey2kind[k]
        self.name = n[0]


def pipe_spec_from_args(args):
    '''
    Makes a pipe spec object from command line arguments
    '''

    node_stack = []
    sequence = None
    node = None
    kkey = None

    pipe = None

    def _open_seq(stack, sname, seq):
        stack.append({ sname: [] })
        if seq is not None:
            seq.append(stack[-1])
        return node_stack[-1][sname]


    def _close_seq():
        node_stack.pop()
        node = node_stack[-1]
        assert len(node) == 1
        k, seq = next(iter(node_stack[-1].items()))
        assert k in ('fanout', 'chain')
        return seq

    sequence = _open_seq(node_stack, 'chain', None)
        
    for a in args:

        if a in ('?', '-?'):
            raise PipeHelpRequested(node)
        
        if a[0:2] == '--':
            # Starting a new node; open a sequence if necessary
            kind, name = _type_spec_from_arg(a[2:])
            #print(f'{kind}.{name} for {sequence}')
            sequence.append({ kind: [ name ] })
            kkey = kind
            node = sequence[-1]
            
        elif a == '{':
            sequence = _open_seq(node_stack, 'fanout', sequence)

        elif a == '[':
            sequence = _open_seq(node_stack, 'chain', sequence)
            
        elif a in ('}', ']'):
            sequence = _close_seq()
            
        else:
            # Regular node argument (positional or keyword)
            #print(a, 'for', node)
            #print('param:', a)
            eq = a.find('=')
            if eq >= 0:
                # Keyword argument
                prm = node.setdefault('params', dict())
                ts = (int, float, dict_from_str, strip_str)
                val = opcast(a[eq+1:], type_sequence=ts)
                #print(a[eq+1:], '->', val, type(val))
                #print(a, ':', a[:eq], '->', a[eq+1:], val, type(val), eq)
                prm[a[:eq]] = val
            else:
                # Positional argument
                node[kkey].append(opcast(a))

    return node_stack[0]


class PipeMaker:
    '''
    Makes a processing pipe out of a valid object representation
    '''
    
    def __init__(self):

        # The type of the current sequence / the sequence that we need
        # to enter on the next __enter__
        self._flow_type = None

        # We always push to the last item of the stack.
        # Initially, this is self._pipes. As we descend into
        # the YAML tree, we add new sequences to the stack, and
        # pop them on every __exit__ (so that the parent sequence
        # captures processors again).
        self._processor_collection_stack = [ (ChainFlow, []) ]

    def __enter__(self, *a, **kw):
        self._processor_collection_stack.append((self._flow_type, []))
        return self


    def __exit__(self, *args):
        level = self._processor_collection_stack.pop()
        flw = level[0](*(level[1]))
        self._processor_collection_stack[-1][-1].append(flw)


    def __call__(self, *args, flow=None):
        ## Args is either sequence-type, or (kind, name, spec)

        if len(args) == 1:
            sequence = args[0]
            self._flow_type = make_pipe_type('flow', sequence)
            return self

        elif len(args) == 3:
            kind, name, spec = args

        # Differentiate between [ "pipe", "arg", ... ]
        # and "pipe": [ "arg", ... ] type of calls.
        a = None
        kw = None
        if isinstance(name, str):
            if isinstance(spec, list):
                a = spec if spec is not None else []
                kw = {}
            elif isinstance(spec, dict):
                a = []
                kw = spec if spec is not None else {}
            elif spec is None:
                a = []
                kw = {}
            else:
                raise RuntimeError(f'{kind}.{name}: invalid args {spec} (of type {type(spec)})')
        else:
            n = name[0]
            a = name[1:]
            name = n
            kw = spec if spec is not None else {}

        astr = ', '.join([f'{k}' for k in a])
        kwstr = ', '.join([f'{k}={v}' for k,v in kw.items()])
        #print(f'{kind}.{name}({astr}, {kwstr})')

        self._processor_collection_stack[-1][-1].append(
            make_pipe_type(kind, name)(*a, **kw)
        )


    @property
    def pipe(self):
        while self._processor_collection_stack:
            level = self._processor_collection_stack.pop()
            self._flw = level[0](*(level[1]))
        return self._flw
            


    def __repr__(self):
        return pformat(self.pipe)


def with_pipe(spec, func, enter=None, exit=None):
    '''
    Walks a pipeline specification structure and calls
    `func(kind, name, params)` for every processing item
    (source, sink or pipe).

    If `enter` and/or `exit` are specified, they're
    expected to contain references to callables of the
    form `proc(sequence_type)`.
    '''

    for item in spec:
        for k in ('chain', 'fanout'):
            if k not in item:
                continue
            v = item[k]
            if hasattr(func, "__enter__"):
                with func(k) as f:
                    with_pipe(v, f)
            else:
                with_pipe(v, func)

        for k in ('from', 'via', 'to'):
            if k not in item:
                continue

            v = item[k]
            func(_ymlkey2kind[k], v, item.get('params', None))



class SchemaSequenceTooShortError(SchemaError): pass

class SchemaSequenceTooLongError(SchemaError): pass


class Sequence(Schema):
    '''
    Custom validator which checks that a list has exactly the
    number (and type) of items it was specified.
    '''

    class VarArguments(Schema):
        '''
        Catch-all for variable-list arguments.
        '''
        pass


    def __init__(self, *items):
        self._items = items


    def __repr__(self):
        si = ', '.join([str(i) for i in self._items])
        return f'{self.__class__.__name__}({si})'


    def validate(self, data):

        #if len(data) > len(self._items):
        #    raise SchemaSequenceTooLongError("too long")

        ditr = iter(data)
        iitr = iter(self._items)
        out = []
        parity = 0
        cnt = 0
        while True:
            try:
                cnt += 1
                i = next(iitr)
                parity += 1
                d = next(ditr)
                parity -= 1
                if not isinstance(i, self.VarArguments):
                    o = Schema(i).validate(d)
                    out.append(o)
                else:
                    # The rest are variable arguments.
                    var_data = [d] + [x for x in ditr]
                    out += i.validate(var_data)
                    break
            except StopIteration as e:
                if parity != 0:
                    # we have fewer data items than data schema items
                    # this means that all schema items from here on
                    # need to be "optional"
                    while True:
                        if not isinstance(i, Optional):
                            raise SchemaSequenceTooShortError(f"{i} at position {cnt} not in data")
                        try: i = next(iitr)
                        except StopIteration: break
                        cnt += 1
                    break
                else:
                    # If we got here, it means that we ran into an StopIteraion at
                    # next(iitr). This is usually the case when we either have too
                    # much data, or when we're at the legit end of a sequence.
                    try:
                        # Try to increment data, too; if this fails, we're at the legit
                        # end of a sequence. If this works, we have too much data.
                        d = next(ditr)
                        raise SchemaSequenceTooLongError(f"too much data at {cnt}")
                    except StopIteration:
                        break

        return out
                
            

class SchemaUnknownPipeError(SchemaError): pass


class PipeValidator(Schema):
    # There are several nasty things going on here.
    #
    # For one, we need recursive validation: our flow elements
    # "chain" and "fanout" contain a list of what would be pipe
    # elements, or other flow elements! Schema() doesn't seem to
    # have an obvious way to do this, so we're going to do a little
    # hack and actually register the class's .validate() function
    # as an evaluator callable for the contents of the flow elements'
    # list.
    #
    # Then there's the issue with the pipe elements: these may contain
    # a "params" key with a dictionary, but the contents of that
    # dictionary are specific to the actual pipe elemen we're going
    # to use. So in order to _actually_ validate, we need to load
    # the module and ask the pipe element class (either through
    # built-in validation info, or through type hint introspection
    # into the pipe class's .__init__() callable.
    # But in order to do this at "params" evaluation time, we
    # actually need to know which class we're talking about -- which
    # is specified in the previous dict item
    # ("from", "via", "to" -> name).
    #
    # So we actually have two ways of doing this:
    #
    #  - One is by replacing the whole sequence with a callable.
    #    This has the benefit of working with any kind of modules,
    #    even external ones, because the evaluation takes place
    #    at execute-time.
    #
    #  - The other is by generating a list of items from
    #    `camagick.pipe` and actually putting together all of the
    #    pipe elements. This has the advantage of generating a
    #    more-or-less static list (at runtime, though), but it
    #    won't properly validate pipe elements from extrnal modules
    #    unknown to camagick. :-/
    #
    # Pick one.

    def __init__(self, other_pipes=None, example=None, **kwargs):

        self.param_validators = self._builtin_validators()

        super().__init__(
            # FIXME: the first part (validate_processor) is a horrible mix
            # of dynamic validation through function call, and a static
            # validation against a statically-built name/kind dependent
            # scheme ("param_validators").
            # This is because we have a funny format when it comes to
            # mixing positional/keyword parameters.
            # Need to rip that apart ASAP and make it more declarative.
            Or(self._validate_processor, {Or("chain", "fanout"): [self.validate]}),
            **kwargs
        )

    def __repr__(self):
        return 'PipeValidator'


    def _builtin_validators(self):
        # Returns the list of validators for built-in types.
        
        # A processor spec can be of different types:
        #
        # a) only-keyword-parameter tupe:
        #    { "from|via|to": "name"
        #      "params": { "key": value, ...} }
        #
        # b) only-positional-parameter type:
        #    { "from|via|to": "name"
        #      "params": [ value, ... ] }
        #
        # c) short-hand mix positional / keyword parameter:
        #    { "from|via|to": [ "name", value, ... ]
        #      "params": { "key": another_value, ... } }
        #        

        self.processors = {}
        param_validators = {}
        for kind in ("source", "sink", "pipe"):
            self.processors[kind] = find_processors(base=f'camagick.{kind}',
                                                    cls='Processor')
            
            param_validators[kind] = {
                name:self._make_params_validator(Cls, kind, name) \
                for name,Cls in self.processors[kind].items()
            }
            
        return param_validators


    def _kind_and_nameline(self, data):
        # this is the "via": ... line.
        # the ... part can either be a string (name only), or
        # an array of [name, *positional_params].
        name = None
        for k in ("from", "via", "to"):
            if k in data:
                name = data[k]
                break

        if name is None:
            raise SchemaUnknownPipeError('/'.join(data.keys()))
        
        return (_ymlkey2kind[k], name)


    def _validate_processor(self, data):

        kind, nameline = self._kind_and_nameline(data)
        if isinstance(nameline, list):
            name = nameline[0]
            if len(nameline)>1:
                params = nameline[1:]
            else:
                params = {}
        else:
            name = nameline
            params = data.get('params', {})

        try:
            scm = self.param_validators[kind][name]
        except KeyError:
            raise SchemaUnknownPipeError(f'Unknown pipe "{name}" '
                                         f'of kind "{kind}"')

        return {_kind2ymlkey[kind]: name, 'params': scm.validate(params)}


    def _make_params_validator(self, Class, kind, name):
        # This returns either a list, or a dictionary, of Schema components
        # able to validate a parameter list intended for Class().
        #
        # The difficulty is that Unlike in Python, for YAML/JSON we need to
        # settle either _lists_ or _dict_ -- no mixing.
        # This means that we either have positional only, or keyword-only
        # arguments (in both cases also considering optional arguments).
        #
        # We need to determine:
        #
        #  - is either possible (dict or list), or is only one allowed?
        #    -> Any POSITIONAL_ONLY present in the annotations without default
        #       value means automatically that we need to go for a list;
        #    -> Same with KEYWORD_ONLY without default value -> need dict.
        #       (If both are the case, we raise an error.)
        #  
        # - are there mandatory arguments (of any kind)? 
        #    -> Generally, all arguments without a default value are mandatory (right?)        
        
        sig = inspect.signature(Class).parameters

        # Lists of argument spec tuples (name, obj)

        # These can be only positional -- mandatory arguments.
        pos_only = [i for i in filter(lambda x: x[1].kind in \
                                      (x[1].POSITIONAL_ONLY, x[1].VAR_POSITIONAL),
                                      sig.items())]

        # These can be only keyword-based -> optional arguments.
        key_only = [i for i in filter(lambda x: \
                                      x[1].kind in (x[1].KEYWORD_ONLY, x[1].VAR_KEYWORD),
                                      sig.items())]

        # Keywords or positional -> optional, but must be listed (as optional)
        # in the accepted list of positional arguments, too.
        pok      = [i for i in filter(lambda x: x[1].kind == x[1].POSITIONAL_OR_KEYWORD, sig.items())]

        # Positional, Mandatory: all positional arguments, no default value
        pos_mandatory = \
            [i for i in pos_only] + \
            [i for i in filter(lambda x: x[1].default == x[1].empty, pok)]

        # Positional, Optional: all other positional arguments (are there any?),
        # and all key-or-positional arguments which have a default value.
        pos_optional = \
            [i for i in filter(lambda x: x[1].default != x[1].empty, pos_only)] + \
            [i for i in filter(lambda x: x[1].default != x[1].empty, pok)]

        # Keyword, Mandatory: are there any _mandatory_ keyword arguments?...
        key_mandatory = []

        # Keyword, Optional: all pok, and all key arguments
        key_optional = \
            [i for i in filter(lambda x: x[1].default != x[1].empty, pok)] + \
            [(str, i[1]) for i in key_only]

        # Returns a usable type (either the annotation, if it is a single type; or 'object')
        # Ideally we'd like to return the individual types of the annotation UnionType here,
        # but I don't know of any elegant way of doing that :-(
        determine_type = \
            lambda x: ((x.annotation if x.annotation is not x.empty else object) \
            if (typing.get_origin(x.annotation) is None) else object) \
               if (x.kind != x.VAR_POSITIONAL) else Sequence.VarArguments([object])


        positional = \
            [ determine_type(pspec)           for pname,pspec in pos_mandatory  ] + \
            [ Optional(determine_type(pspec)) for pname,pspec in pos_optional ]

        keyword = { pname:determine_type(pspec) for pname,pspec in key_mandatory+pos_mandatory }
        keyword.update({ Optional(pname):determine_type(pspec) \
                         for pname,pspec in key_optional  })

        #print(f'{kind}.{name}: Positional:', pformat(positional))
        #print(f'{kind}.{name}: Keyword:', pformat(keyword))

        return Or(Sequence(*positional), keyword)



from camagick.processor import ProcessorBase, SinkBase
import numpy as np
from caproto.asyncio.client import Context
from caproto import ChannelData
from contextlib import suppress

import logging, asyncio

logger = logging.getLogger(__name__)


class Processor(SinkBase):
    def __init__(self, **targets):
        '''
        Writes data to existing EPICS PVs.
        
        Args:
            **targets: key-value pairs of data tags -> full PVs
              where to write each of the datasets.
        '''
        self._targets = targets


    async def startup(self):
        self._ctx = Context()
        tmp = await self._ctx.get_pvs(*[v for k,v in self._targets.items()])
        self._pvs = { k:v for k,v in zip(self._targets,tmp) }
        self._errors = { k:None for k in self._targets }


    async def shutdown(self):
        await self._ctx.disconnect()


    def _maybe_log_error(self, tag, msg):
        e = self._errors.get(tag, None)
        if e is None:
            logger.error(f'msg="Data error, future erros will be silenced" '
                         f'detail="{msg}" tag={tag}')
        self._errors[tag] = e


    def _maybe_clear_error(self, tag):
        if len(self._errors):
            self._errors[tag] = None
            logger.info(f'msg="Available again" tag={tag} pv={self._targets[tag]}')
        

    async def __call__(self, data=None, context=None):
        '''
        Writes data to their corresponding PVs if there's
        a PV waiting.
        '''

        set_list = []

        for tag,pv in self._pvs.items():
            val = data.get(tag, None)
            if val is None:
                self._maybe_log_error(tag=tag, msg=f'Tag not available')
                allowed_tags = ','.join([k for k in data])
                self._maybe_log_error(tag=tag, msg=f"Options: {allowed_tags}")
                continue
            set_list.append(pv.write(val))

        exc = await asyncio.gather(*set_list, return_exceptions=True)

        for t,e in zip(self._pvs, exc):
            if isinstance(e, Exception):
                self._maybe_log_error(t, e)
            else:
                self._maybe_clear_error(t)

        return {
            k:v for k,v in filter(lambda x: x[0] not in self._pvs,
                                  data.items())
        }




from camagick.processor import ProcessorBase

import logging, time, asyncio

logger = logging.getLogger(__name__)

class Processor(ProcessorBase):

    def __init__(self, Hz: float = None, ms: float = None, s: float = None):
        '''
        Slows down processing speed by the specified rate or period length.
        '''
        if Hz is not None:
            self._period = 1.0/Hz
        elif ms is not None:
            self._period = ms / 1000.0
        elif s is not None:
            self._period = s

        self._last_run = 0


    async def startup(self):
        pass


    async def shutdown(self):
        pass


    async def __call__(self, data=None, context=None):

        diff = (time.time()-self._last_run)
        
        if self._period > diff:
            await asyncio.sleep(self._period-diff)

        self._last_run = time.time()
            
        return data




from camagick.processor import ProcessorBase

import logging

class Processor(ProcessorBase):
    '''
    Demotes the specified data labels into the context status.
    '''    
    def __init__(self, *labels):
        '''
        Args:
            *labels: explicit list of data labels to demote (case-sensitive)
        '''
        self._labels = labels
    
    async def startup(self):
        pass

    async def shutdown(self):
        pass
    
    async def __call__(self, data=None, context=None):

        lbl = self._labels if (self._labels is not None and len(self._labels)>0) \
            else data.keys()
        
        data_without_labels = { k:data[k] for k in filter(lambda x: x not in lbl, data) }

        context.update({ k:data[k] for k in filter(lambda x: x in lbl, data) })

        return data_without_labels



#!/usr/bin/python3

import argparse, os, asyncio, importlib

#from camagick.stash import ZarrStash
#from camagick.helpers import FpsCounter
#from camagick.processor import ProcessorBase

#class CAMagickApplication(ProcessorBase):
#    '''
#    Base class from which all CAMagick applications are derived.
#    Implements access to the data stash and the seed/crop history.
#    (This is a bit of overlap with the Processor base class -- needs
#    refactoring.)
#    '''
#
#    pass

    #def __init__(self, args=None, env=None, name=None,
    #             description="CAMagick Application", init=True):
    #    
    #    self._args = args if args is not None else sys.argv
    #    self._env = env if env is not None else os.environ.copy()
    #    self.applicationName = name or os.path.basename(self._args[0])
    #    
    #    self.argparser = self._create_base_argparser(description)

        
    #async def init_base(self, args):
    #    self.stash = ZarrStash(args.stash, lockKey=args.lock)
    #    self.fps = FpsCounter(100)


    #def _create_base_argparser(self, description):
    #    # Create argparser, add basic functionality arguments
    #    # (mostly just data exchange)
    #    #foo = argparse.ArgumentParser()
    #    foo = argparse.ArgumentParser(prog=self.applicationName,
    #                                  description=description,
    #                                  add_help=False)        
    #    foo.add_argument('--stash', action='store',
    #                     help='File path for the data stash')
    #    foo.add_argument('--lock', action='store',
    #                     help='Lock key to use for synchronising access to data')
    #
    #    return foo
    #
    #
    #def run(self):
    #    asyncio.run(self.async_run())



from camagick.processor import ProcessorBase

class Processor(ProcessorBase):
    '''
    Similar to `--only .*`, i.e. passes on all data tags unfiltered.
    '''
    async def __call__(self, data=None, context=None):
        return data
    



import pytest, asyncio, time, traceback
import multiprocessing as mp

from camagick.source.scratch import Processor as GenSource
from camagick.source.ioc import Processor as IocSource
from camagick.source.epics import Processor as EpicsSource

from camagick.sink.ioc import Processor as IocSink
from camagick.sink.epics import Processor as EpicsSink
from camagick.sink.summary import Processor as SummarySink

from camagick.sink.zarr import Processor as ZarrSink

from camagick.flow.chain import Processor as ChainFlow
from camagick.flow.fanout import Processor as FanoutFlow

from camagick.pipe.pace import Processor as Pace
from camagick.pipe.only import Processor as Only
from camagick.pipe.exclude import Processor as Exclude

from camagick.executor import PipeExecutor

import tempfile, os, random, zarr
import numpy as np

@pytest.fixture(scope='function')
def ztmp():
    tmp = tempfile.NamedTemporaryFile(delete=False)
    tmp.close()
    os.unlink(tmp.name)
    yield tmp.name
    #os.unlink(tmp.name)
    pass

async def test_z_append(ztmp):

    print(f"Zarr: {ztmp}")

    z_path = f'{ztmp}#group'+'{gi}/{tag}{di}'

    sink = ZarrSink(z_path, index='idx', mt=False)

    nr_pts = 20
    scalar = np.random.random((nr_pts,))
    array = np.random.random((nr_pts, 23))

    for i in range(nr_pts):
        await sink(
            data={
                'scalar': scalar[i],
                'array':  array[i]
            },
            context={
                'gi': i % 2,
                'di': (int(i/10)+1)*10,
                'idx': int(i/2) % 5
            })

    zf = zarr.open_group(f'{ztmp}', mode='r')
    print('groups:', [g for g in zf.groups()])

    zf0 = zarr.open_group(f'{ztmp}', path='group0', mode='r')
    zf1 = zarr.open_group(f'{ztmp}', path='group1', mode='r')

    print('groups:', [g for g in zf0.groups()] + [g for g in zf1.groups()])
    print('arrays:', [a for a in zf0.arrays()])
    print('arrays:', [a for a in zf1.arrays()])

    assert (zf0['scalar10'][()] == scalar[0:10:2]).all()
    assert (zf0['array10'][()]  ==  array[0:10:2]).all()
    assert (zf0['scalar20'][()] == scalar[10::2]).all()
    assert (zf0['array20'][()]  ==  array[10::2]).all()
    
    assert (zf1['scalar10'][()] == scalar[1:10:2]).all()
    assert (zf1['array10'][()]  ==  array[1:10:2]).all()
    assert (zf1['scalar20'][()] == scalar[11::2]).all()
    assert (zf1['array20'][()]  ==  array[11::2]).all()


async def test_z_overwrite(ztmp):
    print(f"Zarr: {ztmp}")

    sink = ZarrSink(ztmp, index='idx', mode='o', mt=False)

    nr_pts = 20
    half_pts = int(nr_pts/2)
    array = np.array([[i+j*10 for i in range(6)] for j in range(nr_pts)])

    for i in range(half_pts):
        await sink(
            data={ 'array': array[i] },
            context={ 'idx': i })

    for i in range(half_pts):
        await sink(
            data={ 'array': array[i+half_pts] },
            context={ 'idx': i })

    zf = zarr.open_group(f'{ztmp}', mode='r')
    print('groups:', [g for g in zf.groups()])
    print('arrays:', [a for a in zf.arrays()])
    
    assert (zf['array'][()] == array[half_pts:]).all()

    # since we didn't specify the top-up '+' flag,
    # writing beyond the end of the dataset should raise
    with pytest.raises(RuntimeError):
        await sink(data={'array': array[0]},
                   context={'idx': half_pts+2})


async def test_z_reset(ztmp):
    print(f"Zarr: {ztmp}")

    sink = ZarrSink(ztmp, index='idx', mode='x', mt=False)

    nr_pts = 20
    half_pts = int(nr_pts/2)
    array = np.array([[i+j*10 for i in range(6)] for j in range(nr_pts)])

    for i in range(half_pts):
        await sink(
            data={ 'array': array[i] },
            context={ 'idx': i })

    for i in range(5):
        await sink(
            data={ 'array': array[half_pts+i] },
            context={ 'idx': i })
        

    zf = zarr.open_group(f'{ztmp}/', mode='r')
    print('groups:', [g for g in zf.groups()])
    print('arrays:', [a for a in zf.arrays()])    
    assert zf['array'].shape[0] == 5
    assert (zf['array'][()] == array[half_pts:half_pts+5]).all()
    

    # since we didn't specify the top-up '+' flag,
    # writing beyond the end of the dataset should raise
    with pytest.raises(RuntimeError):
        await sink(data={'array': array[0]},
                   context={'idx': half_pts+2})



import asyncio, pytest, time
import tango.asyncio as tas

from tango import DevState, GreenMode
from tango.server import Device, command, attribute
from tango.databaseds.database import main as srv_main

import multiprocessing as mp

from camagick.source.tango import Processor as TangoSource
from camagick.sink.tango import Processor as TangoSink

class AsyncioDevice(Device):
    green_mode = GreenMode.Asyncio

    async def init_device(self):
        await super().init_device()
        self.set_state(DevState.ON)

    @attribute
    async def detail(self):
        return 3.14


@pytest.fixture(scope='function')
def tsrv():
    host = '127.0.0.1'
    port = '10000'
    p = mp.Process(target=srv_main, args=[(
        '2', '-host', host,'-port', port,
    )])
    p.daemon = True
    p.start()
    time.sleep(1.0)
    yield (host, port)
    p.terminate()


def dev_main(inst, host, port):
    from os import environ
    environ['TANGO_HOST'] = f'{host}:{port}'
    AsyncioDevice.run_server(inst)


@pytest.fixture(scope='function')
async def tdev(tsrv):
    inst = '1' #devices/caspytest/1'
    args = [[f'{inst}'], *tsrv]
    print('Tango server running @', tsrv, args)
    p = mp.Process(target=dev_main, args=args)
    p.daemon = True
    p.start()
    yield inst, *tsrv
    p.terminate()
    

async def _test_tdev(tdev):
    print(f'Connecting to Tango test device {tdev}')

    import os
    old_tango_host = os.environ.get('TANGO_HOST', None)

    try:

        os.environ['TANGO_HOST'] = f'{tdev[1]}:{tdev[2]}'
        print('Tango host env-var:', os.environ['TANGO_HOST'])        
        
        src = TangoSource(dev=f'AsyncioDevice/{tdev[0]}', detail='detail')
        await src.startup()
        data = await src()
        print(f'Received: {data}')
        #inst = '3'
        #AsyncioDevice.run_server(inst) #, '-host', '127.0.0.1', '-port', '10000')

        await asyncio.sleep(10.0)
        await src.teardown()

    finally:
        if old_tango_host is not None:
            os.environ['TANGO_HOST'] = old_tango_host




from camagick.processor import ProcessorBase

import logging

class Processor(ProcessorBase):
    '''
    Promotes the specified context labels into the data status.
    '''    
    def __init__(self, *labels):
        '''
        Args:
            *labels: explicit list of data labels to demote (case-sensitive)
        '''
        self._labels = labels
    
    async def startup(self):
        pass

    async def shutdown(self):
        pass
    
    async def __call__(self, data=None, context=None):

        lbl = self._labels if (self._labels is not None and len(self._labels)>0) \
            else context.keys()


        nd = data.copy() if data is not None else {}
        for k in lbl:
            try:
                nd[k] = context[k]
                del context[k]
            except KeyError:
                pass

        return nd



#!/usr/bin/python3

import logging, time, math, random, asyncio

from camagick.processor import ProcessorBase, SourceBase
from camagick.probe import opcast
from numpy import random as np_rand
from numpy import array, zeros, ones

import numpy as np

logger = logging.getLogger(__name__)

class Processor(SourceBase):
    '''
    Produces data from scratch by various means.
    '''

    def __init__(self, method: str, *scalars: str, rate: float = 100.0, opts: dict = None, **arrays):
        '''
        Produces data.

        Args:
            method: string designating the data-producer method.
              currently the following are supported:
                - "sinus" (start, stop, ampl, freq)
                - "random"
                - "zeros"
                - "ones" (factor)
                - "linspace" (start, stop, endpoint)
                - "counter" (start, inc)

            *scalars: list of data tags for scalar values

            rate: Call rate limiter in Hz.

            **arrays: tag -> length map for non-scalar values
        '''

        self._method = getattr(self, f'_make_{method}')
        self._scalars = scalars
        self._arrays = arrays
        self._opts = opts if opts is not None else {}

        self._last_time = 0.0
        self._period = 1.0 / rate

        super().__init__()


    def _make_random(self, shape, **kw):
        return np_rand.random(shape)

    def _make_sinus(self, shape, start=0, stop=6.28, ampl=1.0, freq=1.0):
        return ampl*np.sin(np.linspace(start, stop, shape)*freq)

    def _make_ones(self, shape, factor=1.0):
        return ones(shape)*factor

    def _make_zeros(self, shape, **kw):
        return zeros(shape)

    def _make_linspace(self, shape, start=0.0, stop=1.0, endpoint=True):
        if hasattr(shape, "__len__"):
            if len(shape)>1:
                raise RuntimeError(f'msg="Bad shape for linspace" shape={shape}')
            else:
                num = shape[0]
        else:
            num = shape
        return np.linspace(start, stop, num=num, retstep=False, endpoint=endpoint)

    def _make_counter(self, shape, start=0, inc=1):
        if not hasattr(self, "_cntr"):
            self._cntr = start
        else:
            self._cntr += inc
        return np.ones(shape) * self._cntr


    async def __call__(self, data=None, context=None):

        dwell = min(time.time()-self._last_time, self._period)
        await asyncio.sleep(dwell)
        
        d = {}

        d.update({
            k:self._method(v, **(self._opts)) \
            for k,v in self._arrays.items()
        })

        d.update({
            k:array(self._method([1], **(self._opts))) \
            for k in self._scalars
        })

        return d
